# Copyright © 2021 by Northwestern University.
# 
# LVGP-PyTorch is copyrighted by Northwestern University. It may be freely used 
# for educational and research purposes by  non-profit institutions and US government 
# agencies only. All other organizations may use LVGP-PyTorch for evaluation purposes 
# only, and any further uses will require prior written approval. This software may 
# not be sold or redistributed without prior written approval. Copies of the software 
# may be made by a user provided that copies are not sold or distributed, and provided 
# that copies are used under the same terms and conditions as agreed to in this 
# paragraph.
# 
# As research software, this code is provided on an "as is'' basis without warranty of 
# any kind, either expressed or implied. The downloading, or executing any part of this 
# software constitutes an implicit agreement to these terms. These terms and conditions 
# are subject to change at any time without prior notice.

import numpy as np
import io
from typing import Optional,Union,List,Tuple,Callable

class Variable(object):
    def __init__(self,name:str) -> None:
        self.name = name
    
    def __repr__(self):
        raise NotImplementedError
    
    def sample(self,random_state:np.random.RandomState,size:int):
        '''
        Sample variable uniformly at random.
        '''
        raise NotImplementedError
    
    def _transform(self,vector:np.ndarray):
        raise NotImplementedError

    def _transform_scalar(self,x):
        return self._transform(x)

    def _inverse_transform(self,vector:np.ndarray):
        raise NotImplementedError

    def _inverse_transform_scalar(self,x):
        return self._inverse_transform(x)

class NumericalVariable(Variable):
    """Numerical variable with lower and upper bounds

    :param name: Name of the variable
    :type name: str
    :param lower: Lower bound
    :type lower: Union[int,float]
    :param upper: Upper bound
    :type upper: Union[int,float]
    :param log: If `True`, the variable is sampled on a logarthmic scale (for both
        random and latin-hypercube sampling). Defaults to `False`.
    :type log: bool
    """
    def __init__(
        self, name:str,lower:Union[int,float],
        upper:Union[int,float],
        log:bool=False,
    )-> None:
        super().__init__(name=name)

        self.lower = float(lower)
        self.upper = float(upper)
        self.log = log

        if self.lower >= self.upper:
            raise ValueError(
                'Upper bound must be larger than lower bound for %s' % name
            )
        elif log and self.lower <= 0:
            raise ValueError(
                'Lower bound needs to be positive for log-scale'
            )

        if self.log:
            self._lower = np.log(self.lower)
            self._upper = np.log(self.upper)
        else:
            self._lower,self._upper = self.lower,self.upper
        
    def __repr__(self) -> str:
        repr_str = io.StringIO()
        repr_str.write(
            "%s, Type: Numerical, Range: [%s,%s]" %(
                self.name,repr(self.lower),repr(self.upper))
        )
        if self.log:
            repr_str.write(', on log-scale')
        
        repr_str.seek(0)
        return repr_str.getvalue()
    
    def sample(
        self,random_state:np.random.RandomState,size:int
    ) -> np.ndarray:
        return random_state.uniform(size=size)

    def _transform(self, vector:np.ndarray)->np.ndarray:
        out = self._lower + (self._upper-self._lower)*vector
        if self.log:
            out = np.exp(out)
        
        return np.clip(out,self.lower,self.upper)
    
    def _inverse_transform(self,vector:np.ndarray)->np.ndarray:
        if self.log:
            vector = np.log(vector)
        
        vector = (vector-self._lower)/(self._upper-self._lower)
        return np.clip(vector,0.,1.)

class IntegerVariable(Variable):
    """Integer variable with lower and upper bounds

    The only difference with :obj:`~.NumericalVariable` is that the samples are rounded
    to the closest integer.

    :param name: Name of the variable
    :type name: str
    :param lower: Lower bound
    :type lower: Union[int,float]
    :param upper: Upper bound
    :type upper: Union[int,float]
    :param log: If `True`, the variable is sampled on a logarthmic scale (for both
        random and latin-hypercube sampling). Defaults to `False`.
    :type log: bool
    """
    def __init__(
        self, name:str,lower:int,
        upper:int,
        log:bool=False,
    )-> None:
        super().__init__(name=name)
        self.lower = lower
        self.upper = upper
        self.log = log
        self.numvar = NumericalVariable(name,self.lower,self.upper,self.log)
        
    def __repr__(self) -> str:
        repr_str = io.StringIO()
        repr_str.write(
            "%s, Type: Integer, Range: [%s,%s]" %(
                self.name,repr(self.lower),repr(self.upper))
        )
        if self.log:
            repr_str.write(', on log-scale')
        
        repr_str.seek(0)
        return repr_str.getvalue()
    
    def sample(
        self,random_state:np.random.RandomState,size:int
    ) -> np.ndarray:
        # first sample from [0,1]
        out = self.numvar.sample(random_state,size)
        # then convert back to the original scale
        # and rounding floats if any
        out = self._transform(out)
        # and then inverse-transform back to [0,1]
        return self._inverse_transform(out)

    def _transform(self, vector:np.ndarray)->np.ndarray:
        # transform from [0,1] to the correct scale
        out = self.numvar._transform(vector)
        return np.rint(out).astype(int)
    
    def _inverse_transform(self,vector:np.ndarray)->np.ndarray:
        return self.numvar._inverse_transform(vector)

class CategoricalVariable(Variable):
    '''
    Categorical variable with pre-defined `levels`

    :param name: Name of the variable
    :type name: str
    :param levels: List/tuple of levels. Each level is typically  one of 
        `str`, `float` or `int`.
    :type levels: Union[List,Tuple]
    '''
    def __init__(
        self,name:str,levels:Union[List,Tuple]
    )-> None:
        
        super().__init__(name)
        self.levels = tuple(levels)
        self.num_levels = len(self.levels)
        self.levels_vector = list(range(self.num_levels))

    def __repr__(self) -> str:
        repr_str = io.StringIO()
        repr_str.write("%s, Type: Categorical, Levels: {" % (self.name))
        for idx, choice in enumerate(self.levels):
            repr_str.write(str(choice))
            if idx < len(self.levels) - 1:
                repr_str.write(", ")
        repr_str.write("}")

        repr_str.seek(0)
        return repr_str.getvalue()
    
    def sample(
        self,random_state:np.random.RandomState,size:int
    ) -> np.ndarray:
        return random_state.choice(self.num_levels,size=size,replace=True)

    def stratified_sample(
        self,random_state:np.random.RandomState,size:int
    ) -> np.ndarray:
        if size<self.num_levels:
            raise ValueError
        
        num_mult = size//self.num_levels
        out = np.array(self.levels_vector*num_mult)
        rem = size-num_mult*self.num_levels
        if rem>0:
            out = np.concatenate([
                out,random_state.choice(self.num_levels,size=rem,replace=False)
            ])
        
        random_state.shuffle(out)
        return out

    def _transform(self, vector):
        return np.array([
            self.levels[int(x)] for x in vector
        ])
    
    def _transform_scalar(self,x:int) -> Union[str,float,int]:
        return self.levels[int(x)]

    def _inverse_transform_scalar(self,x:Union[str,float,int]) -> int:
        return self.levels.index(x)
    
    def _inverse_transform(self,vector:np.ndarray) -> np.ndarray:
        return np.array([
            self.levels.index(x) for x in vector
        ])


# Copyright © 2023, Dr. Bostanabad's research group at the University of California, Irvine.
# 
# GP+ Intellectual Property Notice:
# 
# The software known as GP+ is the proprietary material of Dr. Bostanabad's research group at the University of California, Irvine. 
# Non-profit academic institutions and U.S. government agencies may utilize this software exclusively for educational and research endeavors. 
# All other entities are granted permission for evaluation purposes solely; any additional utilization demands prior written consent from the appropriate authority. 
# The direct sale or redistribution of this software, in any form, without explicit written authorization is strictly prohibited. 
# Users are permitted to make duplicate copies of the software, contingent upon the assurance that no copies are sold or redistributed and they adhere to the stipulated terms herein.
# 
# Being academic research software, GP+ is provided on an "as is" m_gp, devoid of warranties, whether explicit or implicit. 
# The act of downloading or executing any segment of this software inherently signifies compliance with these terms. 
# The developers reserve the right to modify these terms and conditions without prior intimation at any juncture.
import numpy as np
from typing import Dict,List,Optional
import math
from scipy.stats import norm
import matplotlib.pyplot as plt
from tabulate import tabulate
import sobol_seq
import warnings

import torch
from torch import nn
from torch import Tensor 
from torch.nn.parameter import Parameter
from torch.nn import init
import torch.nn.functional as F 

import gpytorch
from gpytorch.constraints import Positive
from gpytorch.means import Mean
from gpytorch.priors import NormalPrior
from gpytorch.distributions import MultivariateNormal

from gpplus.visual.plot_latenth import plot_sep
from gpplus.models.gpregression import GPR
from gpplus import kernels
from gpplus.priors import MollifiedUniformPrior
from gpplus.preprocessing import setlevels
from gpplus.utils import set_seed, data_type_check
from gpplus.optim import fit_model_scipy, fit_model_continuation, fit_model_torch


class GP_Plus(GPR):
    """The GP_Plus class extends Gaussian Processes (GPs) to learn nonlinear and probabilistic manifolds, handle categorical inputs, and more.

    :note: Binary categorical variables should not be treated as qualitative inputs. There is no 
        benefit from applying a latent variable treatment for such variables. Instead, treat them
        as numerical inputs.
    
    :param train_x: The training inputs. This represents the input data.
    :param train_y: The training outputs. This represents the output data.
    :param dtype: The data type of the model and data, which could be float32 or float64 torch tensor.
    :param device: Specifies whether the model will be built on CPU or CUDA if available. Otherwise, it should always be CPU.
    :param qual_dict: A dictionary that indicates categorical inputs. The keys represent the number of columns that are categorical, starting from 0, and the values indicate the number of levels for each specific categorical input.
    :param multiple_noise: Used in cases where there are multiple sources of data (multiple fidelity emulation), allowing for different noise estimation for each fidelity data source.
    :param lb_noise: The lower bound for noise estimation.
    :param fix_noise: Determines whether to fix noise to a specific value. It can be True or False. If True, it indicates that noise should be fixed.
    :param fix_noise_val: The value to which noise is fixed when `fix_noise` is True.
    :param quant_correlation_class: The class of correlation used in the model.
    :param fixed_length_scale: Indicates whether the length scale should be fixed. It can be True or False. If True, the length scale will be fixed.
    :param fixed_length_scale_val: The value of the length scale in fixed scenarios.
    :param encoding_type: The type of encoding used for categorical inputs.
    :param embedding_dim: The dimension of the encoding (dimension of embedding).
    :param separate_embedding: Specifies the layers in the embedding used for mapping.
    :param embedding_type: The type of embedding, could be 'deterministic' or 'probabilistic'.
    :param NN_layers_embedding: A list that shows the architecture of neural network layers for the embedding.
    :param m_gp: Specifies the type of mean function considered for the GP model.
    :param m_gp_ref: The mean function for the reference source, which has an ID equal to 0.
    :param NN_layers_m_gp: A list that shows the architecture of the neural network used in the GP model for cases where a neural network is used as the mean function.
    :param calibration_type: The type of calibration used, could be 'deterministic' or 'probabilistic'.
    :param calibration_id: A list that indicates the inputs that have calibration parameters.
    :param mean_prior_cal: The mean prior used for calibration, which should have the same length as `calibration_id`.
    :param std_prior_cal: The standard deviation prior used for calibration, which should have the same length as `calibration_id`.
    :param interval_score: Indicates whether to add interval scoring during optimization.It can be True or False. If True, it is added by default with a coefficient of 0.08.
    :param seed_number: The seed number for random number generation to ensure reproducible results.
    """
    def __init__(
        self,
        train_x:torch.Tensor,
        train_y:torch.Tensor,
        dtype= torch.float,
        device="cpu",
        qual_dict = {},
        multiple_noise = False,
        lb_noise:float=1e-8,
        fix_noise:bool=False,
        fix_noise_val:float=1e-5,
        quant_correlation_class:str='Rough_RBF',
        fixed_length_scale:bool=False,
        fixed_length_scale_val=torch.tensor([1.0]),
        encoding_type = 'one-hot',
        embedding_dim:int=2,
        separate_embedding = [] , #max=2
        embedding_type='deterministic',
        NN_layers_embedding:list = [],
        m_gp='single_constant',
        m_gp_ref='zero',
        NN_layers_m_gp=[],
        calibration_type='deterministic',
        calibration_id=[],
        mean_prior_cal=None,  
        std_prior_cal=None,  
        interval_score=False,
        num_pass_train=1,
        num_pass_pred=1,
        seed_number=1
    ) -> None:

        
        if mean_prior_cal is None:
            mean_prior_cal = [0 for _ in calibration_id]
        if std_prior_cal is None:
            std_prior_cal = [1 for _ in calibration_id]
        
        self.mean_prior_cal=mean_prior_cal
        self.std_prior_cal=std_prior_cal
        
        self.interval_score=interval_score
        tkwargs = {}  # or dict()
        tkwargs['dtype'] = dtype
        tkwargs['device'] =torch.device(device)
        self.tkwargs=tkwargs


        self.mean_prior_cal=mean_prior_cal
        self.std_prior_cal=std_prior_cal
        if fixed_length_scale:
            self.fixed_length_scale_val=fixed_length_scale_val.to(**self.tkwargs)
        else:
            self.fixed_length_scale_val=None
        
        if calibration_type=='probabelistic' or  calibration_type=='probabelistic':
            num_pass_train=20
            num_pass_pred=30
        # Checking inputs & output
        train_x=data_type_check(train_x)
        train_y=data_type_check(train_y)

        if not isinstance(qual_dict, dict):
            raise ValueError("qual_dict should be a dictionary.") 

        if multiple_noise not in [True, False]:
            raise ValueError("multiple_noise should be either True or False.")

        if not isinstance(embedding_dim, int):
            raise ValueError("embedding_dim should be an integer.")

        if quant_correlation_class not in ['Rough_RBF', 'RBFKernel', 'Matern32Kernel', 'Matern12Kernel','Matern52Kernel']:
            raise ValueError("quant_correlation_class should be 'Rough_RBF', 'RBFKernel', 'Matern32Kernel', 'Matern12Kernel','Matern52Kernel'.")

        if fix_noise not in [True, False]:
            raise ValueError("fix_noise should be either True or False.")

        if not isinstance(NN_layers_embedding, list) or not all(isinstance(i, int) for i in NN_layers_embedding):
            raise ValueError("NN_layers_embedding should be a list of integers representing the number of neurons in each layer.")

        if encoding_type != 'one-hot':
            raise ValueError("encoding_type should be 'one-hot'.")

        if embedding_type not in ['deterministic', 'probabilistic']:
            raise ValueError("embedding_type should be either 'deterministic' or 'probabilistic'.")

        if not isinstance(separate_embedding, list) or not all(isinstance(i, int) for i in separate_embedding):
            raise ValueError("separate_embedding should be a list with integers showing the number of categorical inputs to be considered in a separate manifold in each layer.")

    
        supported_singl_functions = ['single_sin', 'single_cos', 'single_exp', 'single_log', 'single_tan', 'single_asin', 'single_acos', 'single_atan', 
                                    'single_sinh', 'single_cosh', 'single_tanh', 'single_asinh', 'single_acosh', 'single_atanh', 'single_sqrt', 
                                    'single_abs', 'single_ceil', 'single_floor', 'single_round'] ########## single_functions!!!!

        self.supported_singl_m_gp_functions=supported_singl_functions
        
        supported_multi_m_gp_functions=['single_zero', 'single_polynomial', 'single_constant', 'multiple_polynomial_2d', 'multiple_constant', 'neural_network']


        if not isinstance(NN_layers_m_gp, list) or not all(isinstance(i, int) for i in NN_layers_m_gp):
            raise ValueError("NN_layers_m_gp should be a list with integers representing the number of neurons in each layer for the mean function.")

        if not isinstance(calibration_id, list) or not all(isinstance(i, int) for i in calibration_id):
            raise ValueError("calibration_id should be a list where each entry shows the column number in the dataset that the calibration parameters are assigned to.")
        
        train_x=self.fill_nan_with_mean(train_x,calibration_id)
        ###############################################################################################
        ###############################################################################################
        self.seed=seed_number
        self.calibration_id=calibration_id
        self.calibration_source_index=0    ## It is supposed the calibration parameter is for high fidelity needs
        qual_dict_list = list(qual_dict.keys())
        all_index = set(range(train_x.shape[-1]))
        quant_index = list(all_index.difference(qual_dict_list))
        num_levels_per_var = list(qual_dict.values())
        #------------------- lm columns --------------------------
        lm_columns = list(set(qual_dict_list).difference(separate_embedding))
        if len(lm_columns) > 0:
            qual_kernel_columns = [*separate_embedding, lm_columns]
        else:
            qual_kernel_columns = separate_embedding
        #########################
        train_y=train_y.reshape(-1)
        if multiple_noise:
            noise_indices = list(range(0,num_levels_per_var[-1]))
        else:
            noise_indices = []

        if len(qual_dict_list) == 1 and num_levels_per_var[0] < 2:
            temp = quant_index.copy()
            temp.append(qual_dict_list[0])
            quant_index = temp.copy()
            qual_dict_list = []
            embedding_dim = 0
        elif len(qual_dict_list) == 0:
            embedding_dim = 0

        if len(qual_dict_list) == 0:
            embedding_dim = 0

        if len(qual_dict_list) > 0:
            ####################### Defined multiple kernels for seperate variables ###################
            qual_kernels = []
            for i in range(len(qual_kernel_columns)):
                qual_kernels.append(kernels.RBFKernel(
                    active_dims=torch.arange(embedding_dim) + embedding_dim * i) )
                qual_kernels[i].initialize(**{'lengthscale':1.0})
                qual_kernels[i].raw_lengthscale.requires_grad_(False)
                
        quant_correlation_class_name = quant_correlation_class  
        if quant_correlation_class_name == 'Rough_RBF':
            quant_correlation_class = 'RBFKernel'
        if len(quant_index) == 0:
            correlation_kernel = qual_kernels[0]
            for i in range(1, len(qual_kernels)):
                correlation_kernel *= qual_kernels[i]
        else:
            try:
                quant_correlation_class = getattr(kernels,quant_correlation_class)
            except:
                raise RuntimeError(
                    "%s not an allowed kernel" % quant_correlation_class
                )
            if quant_correlation_class_name == 'RBFKernel':
                quant_kernel = quant_correlation_class(
                    ard_num_dims=len(quant_index),
                    active_dims=len(qual_kernel_columns) * embedding_dim+torch.arange(len(quant_index)),
                    lengthscale_constraint= Positive(transform= torch.exp,inv_transform= torch.log)
                )
            elif quant_correlation_class_name == 'Rough_RBF':
                quant_kernel = quant_correlation_class(
                    ard_num_dims=len(quant_index),
                    active_dims=len(qual_kernel_columns)*embedding_dim+torch.arange(len(quant_index)),
                    lengthscale_constraint= Positive(transform= lambda x: 2.0**(-0.5) * torch.pow(10,-x/2),inv_transform= lambda x: -2.0*torch.log10(x/2.0))
                )
            elif quant_correlation_class_name == 'Matern12Kernel':
                quant_kernel = quant_correlation_class(
                    ard_num_dims=len(quant_index),
                    active_dims=len(qual_kernel_columns)*embedding_dim+torch.arange(len(quant_index)),
                    lengthscale_constraint= Positive(transform= lambda x: 2.0**(-0.5) * torch.pow(10,-x/2),inv_transform= lambda x: -2.0*torch.log10(x/2.0))
                )
            
            elif quant_correlation_class_name == 'Matern32Kernel':
                quant_kernel = quant_correlation_class(
                    ard_num_dims=len(quant_index),
                    active_dims=len(qual_kernel_columns)*embedding_dim+torch.arange(len(quant_index)),
                    lengthscale_constraint= Positive(transform= lambda x: 2.0**(-0.5) * torch.pow(10,-x/2),inv_transform= lambda x: -2.0*torch.log10(x/2.0))             
                )
            elif quant_correlation_class_name == 'Matern52Kernel':
                quant_kernel = quant_correlation_class(
                    ard_num_dims=len(quant_index),
                    active_dims=len(qual_kernel_columns)*embedding_dim+torch.arange(len(quant_index)),
                    lengthscale_constraint= Positive(transform= lambda x: 2.0**(-0.5) * torch.pow(10,-x/2),inv_transform= lambda x: -2.0*torch.log10(x/2.0))       
                )
                #####################
            if quant_correlation_class_name == 'RBFKernel':
                quant_kernel.register_prior(
                    'lengthscale_prior', MollifiedUniformPrior(math.log(0.1),math.log(10)),'raw_lengthscale'
                )
                
            elif quant_correlation_class_name == 'Rough_RBF':
                quant_kernel.register_prior(
                    'lengthscale_prior',NormalPrior(-3.0,3.0),'raw_lengthscale'
                )
            elif quant_correlation_class_name == 'Matern12Kernel':
                quant_kernel.register_prior(
                    'lengthscale_prior',NormalPrior(-3.0,3.0),'raw_lengthscale'
                )

            elif quant_correlation_class_name == 'Matern32Kernel':
                quant_kernel.register_prior(
                    'lengthscale_prior',NormalPrior(-3.0,3.0),'raw_lengthscale'
                )

            elif quant_correlation_class_name == 'Matern52Kernel':
                quant_kernel.register_prior(
                    'lengthscale_prior',NormalPrior(-3.0,3.0),'raw_lengthscale'
                )
            if len(qual_dict_list) > 0:
                temp = qual_kernels[0]
                for i in range(1, len(qual_kernels)):
                    temp *= qual_kernels[i]
                correlation_kernel = temp*quant_kernel #+ qual_kernel + quant_kernel
            else:
                correlation_kernel = quant_kernel
            #####################
        super(GP_Plus,self).__init__(
            train_x=train_x,train_y=train_y,noise_indices=noise_indices,
            correlation_kernel=correlation_kernel,
            fix_noise=fix_noise,fix_noise_val=fix_noise_val,lb_noise=lb_noise
        )

        self.calibration_type=calibration_type
        for n, mean_prior, std_prior in zip(self.calibration_id, self.mean_prior_cal, self.std_prior_cal):
            if self.calibration_type == 'probabilistic':
                setattr(self, 'Theta_' + str(n), LinearVariational(batch_shape=torch.Size([]), mean_prior=mean_prior, std_prior=0*std_prior).to(**tkwargs))
                setattr(self, 'calibration_element' + str(n), torch.where(train_x[:, -1] == self.calibration_source_index)[0])
            else:
                setattr(self, 'Theta_' + str(n), gpytorch.means.ConstantMean(prior=NormalPrior(mean_prior, std_prior)))
                setattr(self, 'calibration_element' + str(n), torch.where(train_x[:, -1] == self.calibration_source_index)[0])

            train_x[getattr(self, 'calibration_element' + str(n)), n] = torch.zeros_like(train_x[getattr(self, 'calibration_element' + str(n)), n])

        # register index and transforms
        self.register_buffer('quant_index',torch.tensor(quant_index))
        self.register_buffer('qual_dict_list',torch.tensor(qual_dict_list))

        self.qual_kernel_columns = qual_kernel_columns
        # latent variable mapping
        self.num_levels_per_var = num_levels_per_var
        self.embedding_dim = embedding_dim
        self.encoding_type = encoding_type
        self.embedding_type=embedding_type
        self.perm =[]
        self.zeta = []
        self.random_zeta=[]
        self.perm_dict = []
        self.A_matrix = []
        self.epsilon=None
        self.epsilon_f=None
        self.embeddings_Dtrain=[]
        self.count=train_x.size()[0]
        self.num_pass_train=num_pass_train
        self.num_pass_pred=num_pass_pred
        if len(qual_kernel_columns) > 0:
            for i in range(len(qual_kernel_columns)):
                if type(qual_kernel_columns[i]) == int:
                    num = self.num_levels_per_var[qual_dict_list.index(qual_kernel_columns[i])]
                    cat = [num]
                else:
                    cat = [self.num_levels_per_var[qual_dict_list.index(k)] for k in qual_kernel_columns[i]]
                    num = sum(cat)

                zeta, perm, perm_dict = self.zeta_matrix(num_levels=cat, embedding_dim = self.embedding_dim)
                self.zeta.append(zeta)
                self.perm.append(perm)
                self.perm_dict.append(perm_dict)       
                ###################################  latent map (manifold) #################################   
                if self.embedding_type=='probabilistic':
                    setattr(self,'A_matrix', Variational_Encoder(self, input_size= num, num_classes=5, 
                        layers =NN_layers_embedding, name = str(qual_kernel_columns[i])).to(**tkwargs))
                else:
                    model_temp = FFNN(self, input_size= num, num_classes=embedding_dim, 
                        layers = NN_layers_embedding, name ='latent'+ str(qual_kernel_columns[i])).to(**self.tkwargs)
                    self.A_matrix.append(model_temp)

        ##################################################################################
        if fixed_length_scale == True:
            self.covar_module.base_kernel.raw_lengthscale.data = self.fixed_length_scale_val 
            self.covar_module.base_kernel.raw_lengthscale.requires_grad = False  
        ###################################  Mean Function #################################   
        i=0
        self.m_gp=m_gp
        self.m_gp_ref=m_gp_ref
        self.num_sources=int(torch.max(train_x[:,-1]))
        size=train_x.shape[1]
        if self.m_gp.startswith('single'):
            self.single_m_gp_register(size,m_gp_type=self.m_gp,wm='mean_module')
        elif self.m_gp.startswith('multi'):
            self.multi_m_gp_register (train_x,supported_multi_m_gp_functions,self.m_gp_ref)
        elif self.m_gp=='neural_network': 
            setattr(self,'mean_module_NN_All', FFNN_as_Mean(self, input_size= train_x.shape[1]+2*len(qual_dict_list)-len(qual_dict_list), num_classes=1,layers =NN_layers_m_gp, name = str('mean_module_'+str(i)+'_')).to(**tkwargs)) 
        else: 
            raise ValueError('The "m_gp" argument must start with "multi", "single", or "neural_network".')



    def forward(self,x:torch.Tensor) -> MultivariateNormal:
        if self.embedding_type=='probabilistic' or self.calibration_type=='probabilistic':
            set_seed(self.seed)
            if self.training:
                Numper_of_pass=self.num_pass_train#5 
            else:
                Numper_of_pass=self.num_pass_pred#10 
        else:
            Numper_of_pass=1
        
        Sigma_sum=torch.zeros(x.size(0),x.size(0), dtype=torch.float64).to(self.tkwargs['device'])
        mean_x_sum=torch.zeros(x.size(0), dtype=torch.float64).to(self.tkwargs['device'])

        for NP in range(Numper_of_pass):
            x_forward_raw=x.clone()
            nd_flag = 0
            if x.dim() > 2:
                xsize = x.shape
                x = x.reshape(-1, x.shape[-1])
                nd_flag = 1
            
            x_new= x
            if len(self.qual_kernel_columns) > 0:
                embeddings = []
                for i in range(len(self.qual_kernel_columns)):
                    temp= self.transform_categorical(x=x[:,self.qual_kernel_columns[i]].clone().type(torch.int64).to(self.tkwargs['device']), 
                        perm_dict = self.perm_dict[i], zeta = self.zeta[i])
                dimm=x_forward_raw.size()[0]
                if self.embedding_type=='probabilistic': 
                    # Convert to list of tuples
                    x_raw=torch.zeros(temp.size(0),2)
                    # Find unique rows
                    unique_rows, indices = torch.unique(temp, dim=0, return_inverse=True)
                    temp= unique_rows
                    dimm=unique_rows.size()[0]
                    if self.training:
                        epsilon=torch.normal(mean=0,std=1,size=[dimm,2])
                        embeddings.append(getattr(self,'A_matrix')(x=temp.float().to(**self.tkwargs),epsilon=epsilon))
                    else:
                        if x.size()[0]==self.count:
                            epsilon=torch.normal(mean=0,std=1,size=[dimm,2])
                            embeddings.append(getattr(self,'A_matrix')(x=temp.float().to(**self.tkwargs),epsilon=epsilon))
                            self.embeddings_Dtrain.append(embeddings[0])
                        else:
                            embeddings.append(self.embeddings_Dtrain[NP])
                    for i, index in enumerate(indices):
                        x_raw[i] = embeddings[0][index]
                    embeddings=x_raw
                    x_new= torch.cat([embeddings,x[...,self.quant_index.long()]],dim=-1)
                else:
                    embeddings.append(self.A_matrix[i](temp.float().to(**self.tkwargs)))
                    x_new= torch.cat([embeddings[0],x[...,self.quant_index.long()].to(**self.tkwargs)],dim=-1)
                
                ## For Calibration
                if len(self.calibration_id)>0:
                    if self.training:
                        for n in self.calibration_id:
                            if self.calibration_type=='probabilistic':
                                s=torch.ones_like(x_new[getattr(self,'calibration_element'+str(n)),embeddings[0].size(1)+n]).shape
                                epsilon=torch.normal(mean=0,std=1,size=[s[0],1])
                                Theta=(getattr(self,'Theta_'+str(n))(epsilon.clone().reshape(-1,1)))
                                x_new[getattr(self,'calibration_element'+str(n)),embeddings[0].size(1)+n]=\
                                    torch.ones_like(x_new[getattr(self,'calibration_element'+str(n)),embeddings[0].size(1)+n])*(Theta.reshape(-1))
                                x_new[getattr(self,'calibration_element'+str(n)),embeddings[0].size(1)+n]=\
                                    torch.ones_like(x_new[getattr(self,'calibration_element'+str(n)),embeddings[0].size(1)+n])*(getattr(self,'Theta_'+str(n))(x[i,-1].clone().flatten().reshape(-1,1)))
                    else:
                        for n in self.calibration_id:
                            if self.calibration_type=='probabilistic':
                                epsilon=torch.normal(mean=0,std=1,size=[1,1])
                                calibration_element=torch.where(x[:, -1]==self.calibration_source_index)[0]
                                x_new[calibration_element,embeddings[0].size(1)+n]=\
                                        torch.ones_like(x_new[calibration_element,embeddings[0].size(1)+n])*(getattr(self,'Theta_'+str(n))(epsilon.clone().reshape(-1,1)))
                            else:
                                calibration_element=torch.where(x[:, -1]==self.calibration_source_index)[0]
                                x_new[calibration_element,embeddings[0].size(1)+n]=\
                                    torch.ones_like(x_new[calibration_element,embeddings[0].size(1)+n])*(getattr(self,'Theta_'+str(n))(x[i,-1].clone().reshape(-1,1)))               
            if nd_flag == 1:
                x_new = x_new.reshape(*xsize[:-1], -1)
        #################### Multiple baises (General Case) ####################################  
            if self.m_gp.startswith('multi'):
                mean_x = self.multi_mean(x_new,x_forward_raw).to(**self.tkwargs) 
            elif self.m_gp.startswith('neural_network'):
                mean_x = getattr(self, 'mean_module_NN_All')(x_new.clone().detach()).reshape(-1)
            else:
                mean_x = self.single_mean(x_new).to(**self.tkwargs) 

            covar_x = self.covar_module(x_new).to(**self.tkwargs)
            mean_x_sum+=mean_x
            Sigma_sum += covar_x.evaluate()+ torch.outer(mean_x, mean_x)

        # End of the loop for forward pasess ----> Compute ensemble mean and covariance
        k = Numper_of_pass
        ensemble_mean = mean_x_sum/k
        ensemble_covar = torch.zeros_like(Sigma_sum) 
        ensemble_covar= Sigma_sum/k
        ensemble_covar -= torch.outer(ensemble_mean, ensemble_mean)
        ensemble_covar=gpytorch.lazy.NonLazyTensor(ensemble_covar)
        Sigma_sum=0
        return MultivariateNormal(ensemble_mean,ensemble_covar)
    
    ################################################################ Mean Functions #####################################################################################
    
    def single_m_gp_register (self,size=1,m_gp_type='single_zero',wm='mean_module'):
        if m_gp_type in self.supported_singl_m_gp_functions:
            setattr(self,wm, LinearMean_with_prior(input_size=size, batch_shape=torch.Size([]), bias=False)) 
        elif m_gp_type.startswith('single_polynomial'):
            degree = int(m_gp_type.split('d')[-1])
            setattr(self,wm, LinearMean_with_prior(input_size=degree*(size), batch_shape=torch.Size([]), bias=True)) 
        elif m_gp_type=='single_constant':
            setattr(self,wm, gpytorch.means.ConstantMean(prior=NormalPrior(0.,1)) )
        elif m_gp_type=='single_zero':
            setattr(self,wm, gpytorch.means.ZeroMean())  
    
    def multi_m_gp_register(self,train_x,supported_multi_m_gp_functions,m_gp_ref):
        size=train_x.shape[1]
        if self.m_gp in supported_multi_m_gp_functions:
            for i in range(self.num_sources +1):
                if i==0:
                    m_gp_type= 'single_'+m_gp_ref
                else:
                    m_gp_type='single'+self.m_gp[8:]
                self.single_m_gp_register(size,m_gp_type=m_gp_type,wm='mean_module_'+str(i))

    def single_mean(self, x):
        m_gp_type=self.m_gp
        supported_functions = ['sin', 'cos', 'exp', 'log', 'tan', 'asin', 'acos', 'atan', 
                               'sinh', 'cosh', 'tanh', 'asinh', 'acosh', 'atanh', 'sqrt', 
                               'abs', 'ceil', 'floor', 'round']
        
        if m_gp_type in supported_functions:
            # Dynamically call the PyTorch function m_gpd on m_gp_type
            transformed_x = getattr(torch, m_gp_type)(x.clone()).float()
        elif m_gp_type.startswith('polynomial-d'):
            degree = int(m_gp_type.split('d')[-1])
            transformed_x = x.clone().double()
            polynomial_terms = [transformed_x.pow(n).float() for n in range(1, degree + 1)]
            transformed_x = torch.cat(polynomial_terms, dim=1)
        else:
            # Default case
            transformed_x = x.float().clone()
        mean_x = getattr(self, 'mean_module')(transformed_x)
        return mean_x

    def multi_mean(self,x,x_forward_raw):
        mean_x=torch.zeros_like(x[:,-1])
        if self.m_gp=='multiple_constant':
            for i in range(len(mean_x)):
                qq=int(x_forward_raw[i,-1])                        
                mean_x[i] = getattr(self, 'mean_module_' + str(qq))(x_forward_raw[i,:].clone().float().reshape(1,-1))
        elif self.m_gp=='multiple_polynomial':
            for i in range(len(mean_x)):
                qq=int(x_forward_raw[i,-1])
                # mean_x[i]=getattr(self,'mean_module_'+str(qq))(torch.cat((torch.tensor((x[i,-1].clone().double().reshape(-1,1).float())**2),torch.tensor(x[i,-1].clone().double()).reshape(-1,1).float()),1))
                mean_x[i] = getattr(self, 'mean_module_' + str(qq))(torch.cat((x_forward_raw.clone().detach().double().reshape(-1, 1).float() ** 2,
                        x[i, -1].clone().detach().double().reshape(-1, 1).float()),1))
        
        elif self.m_gp=='neural_network':
            mean_x = getattr(self, 'mean_module_NN_All')(x.clone()).reshape(-1)
        return mean_x 

    ################################################################ Fit #####################################################################################
    def fit(self,add_prior:bool=True,num_restarts:int=64,theta0_list:Optional[List[np.ndarray]]=None,jac:bool=True,
            options:Dict={},n_jobs:int=-1,method = 'L-BFGS-B',constraint=False,bounds=False,regularization_parameter:List[int]=[0,0],optim_type='scipy'):
        print("## Learning the model's parameters has started ##")

        if self.tkwargs['device'].type == 'cuda':
            if optim_type == 'adam_torch':
                fit_model_torch(model=self,
                                model_param_groups=None,
                                lr_default=0.01,
                                num_iter=100,
                                num_restarts=64,
                                break_steps=50)
            else:
                # Issue a warning and proceed with the 'adam_torch' optimizer
                warnings.warn('The model is built to run on CUDA (GPU), but the current optimization type is invalid for this configuration. So, the optimizer is now using adam_torch to train the model.')
                fit_model_torch(model=self.to(**self.tkwargs),
                                model_param_groups=None,
                                lr_default=0.01,
                                num_iter=100,
                                num_restarts=4,
                                break_steps=50)
        else:
            if optim_type=='scipy':
                fit_model_scipy (self,add_prior,num_restarts,theta0_list,jac, options,n_jobs,method ,constraint,bounds,regularization_parameter)
            elif optim_type=='continuation':
                fit_model_continuation(model=self,add_prior=add_prior,
                num_restarts=num_restarts,criterion='NLL',
                initial_noise_var=1,
                red_factor=math.sqrt(10),
                options=options,
                n_jobs= n_jobs,
                accuracy = 1e-2,
                method = method,
                constraint=constraint,
                regularization_parameter=regularization_parameter,
                bounds=bounds
                )
            elif optim_type=='adam_torch':
                fit_model_torch (model=self,
                    model_param_groups=None,
                    lr_default=0.01,
                    num_iter=100,
                    num_restarts=num_restarts,
                    break_steps= 50)
            else:
                raise ValueError(
                    'Invalid optim_type. You must choose one of the following: '
                    '"scipy" (default), "continuation", or "adam_torch".\n'
                    '- "scipy": Uses the SciPy library for optimization, suitable for most CPU-m_gpd computations.\n'
                    '- "continuation": A method designed for more complex optimization scenarios, potentially offering better results in most cases with higher computational cost.\n'
                    '- "adam_torch": Employs the Adam optimizer from the PyTorch library, optimized for GPU-m_gpd computations and large datasets.'
                )
        print("## Learning the model's parameters is successfully finished ##")




    def fill_nan_with_mean(self,train_x,cal_ID):
        # Check if there are any NaNs in the tensor
        if torch.isnan(train_x).any():
            if len(cal_ID)==0:
                print("There are NaN values in the data, which will be filled with column-wise mean values.")
            else:
                print("There are NaN values in the data, which will be estimated in calibration process")
            # Compute the mean of non-NaN elements column-wise
            col_means = torch.nanmean(train_x, dim=0)
            # Find indices where NaNs are located
            nan_indices = torch.isnan(train_x)
            # Replace NaNs with the corresponding column-wise mean
            train_x[nan_indices] = col_means.repeat(train_x.shape[0], 1)[nan_indices]

        return train_x
    ############################  Prediction and Visualization  ###############################
    
    def predict(self, Xtest,return_std=True, include_noise = True):
        Xtest=data_type_check(Xtest)
        with torch.no_grad():
            return super().predict(Xtest, return_std = return_std, include_noise= include_noise)
    
    def predict_with_grad(self, Xtest,return_std=True, include_noise = True):
        Xtest=data_type_check(Xtest)
        return super().predict(Xtest, return_std = return_std, include_noise= include_noise)
    
    def noise_value(self):
        noise = self.likelihood.noise_covar.noise.detach() * self.y_std**2
        return noise

    def score(self, Xtest, ytest, plot_MSE = True, title = None, seperate_levels = False):
        Xtest=data_type_check(Xtest)
        ytest=data_type_check(ytest)
        ytest=ytest.reshape(-1).to(self.tkwargs['device'])
        Xtest=Xtest.to(self.tkwargs['device'])
        plt.rcParams.update({'font.size': 14})
        ypred = self.predict(Xtest, return_std=False)
        mse = ((ytest.reshape(-1)-ypred)**2).mean()
        print('################MSE######################')
        print(f'MSE = {mse:.5f}')
        print('################Noise####################')
        noise = self.likelihood.noise_covar.noise.detach() * self.y_std**2
        
        print(f'The estimated noise parameter (varaince) is {noise}')
        print(f'The estimated noise std is {np.sqrt(noise.cpu())}')
        print('#########################################')

        if plot_MSE:
            _ = plt.figure(figsize=(8,6))
            _ = plt.plot(ytest.cpu().numpy(), ypred.cpu().numpy(), 'ro', label = 'Data')
            _ = plt.plot(ytest.cpu().numpy(), ytest.cpu().numpy(), 'b', label = 'MSE = ' + str(np.round(mse.detach().item(),3)))
            _ = plt.xlabel(r'Y_True')
            _ = plt.ylabel(r'Y_predict')
            _ = plt.legend()
            if title is not None:
                _ = plt.title(title)

        if seperate_levels and len(self.qual_dict_list) > 0:
            for i in range(self.num_levels_per_var[0]):
                index = torch.where(Xtest[:,self.qual_dict_list] == i)[0]
                _ = self.score(Xtest[index,...], ytest[index], 
                    plot_MSE=True, title = 'results' + ' Only Source ' + str(i), seperate_levels=False)
        return ypred

    def plot_xy(self, Xtest, ytest, input_column):
        Xtest=data_type_check(Xtest)
        ytest=data_type_check(ytest)
        
        ytest=ytest.reshape(-1).to(**self.tkwargs)
        Xtest=Xtest.to(**self.tkwargs)
        if len(input_column) > 2:
            raise ValueError("Visualization can only be done for one or two input versions.")

        plt.rcParams.update({'font.size': 14})
        ypred = self.predict(Xtest, return_std=False)

        if len(input_column) == 1:
            plt.scatter(Xtest[:,input_column[0]].cpu().numpy(), ypred.cpu().numpy(),marker= 'o', label='Prediction')
            plt.xlabel('Input: ' + str(input_column[0]))
            plt.ylabel('Output')
        elif len(input_column) == 2:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            ax.scatter(Xtest[:,input_column[0]].cpu().numpy(), Xtest[input_column[1]].cpu().numpy(), ytest.cpu().numpy(), c='r', marker='o', label='Ground Truth')
            ax.scatter(Xtest[:,input_column[0]].cpu().numpy(), Xtest[input_column[1]].cpu().numpy(), ypred.cpu().numpy(), c='b', marker='^', label='Prediction')
            ax.set_xlabel('Input: ' + str(input_column[0]))
            ax.set_ylabel('Input: ' + str(input_column[1]))
            ax.set_zlabel('Output')

        plt.title('Input(s): ' + str(input_column) + ' versus Output')
        plt.legend()
        plt.show()



    def plot_predict_xy(self, var_input, value_range, val_fixed_inputs, num_points=100):
        """
        Plots the prediction of the model for a specified input variable while keeping other inputs fixed.

        Parameters:
        - var_input (int): The index of the input variable to vary.
        - value_range (list or tuple): The range [min, max] over which to vary var_input.
        - val_fixed_inputs (dict): Dictionary with column indices as keys and fixed values as values for all input variables except var_input.
        - num_points (int): The number of points in the linspace for var_input. Default is 100.

        Raises:
        - ValueError: If var_input is not a valid index for the input variables.
        """

        # Check if var_input is a valid index
        if not (0 <= var_input < len(val_fixed_inputs)+1):
            raise ValueError("var_input must be a valid index of the input variables.")

        # Define the number of points in the linspace
        var_input_values = np.linspace(value_range[0], value_range[1], num_points)
        
        # Initialize Xtest with the appropriate shape
        Xtest = np.zeros((num_points, len(val_fixed_inputs)+1))

        # Fill Xtest with fixed values
        for i in range(Xtest.shape[1]):
            if i == var_input:
                Xtest[:, i] = var_input_values
            elif i in val_fixed_inputs:
                Xtest[:, i] = val_fixed_inputs[i]
            else:
                raise ValueError(f"Column index {i} is not specified in val_fixed_inputs and is not the variable input index.")

        Xtest=torch.tensor(Xtest)        
        # Make predictions
        ypred = self.predict(Xtest, return_std=False)
        # Plotting
        plt.plot(Xtest[:, var_input].cpu().numpy(), ypred.cpu().numpy(), label='Prediction')
        plt.xlabel('Input ' + str(var_input))
        plt.ylabel('Output')
        plt.title('Input: ' + str(var_input) + ' versus Output')
        plt.legend()
        plt.show()
    def plot_1d(self, Xtest, ytest, Xtrain, ytrain, model, label_test='$Exact$', 
                         label_prediction='Predicted', Train_label='Training data', 
                         xlabel='x', ylabel='y', ax=None, x_lower=None, x_upper=None, 
                         y_lower=None, y_upper=None):
        Xtest = data_type_check(Xtest)
        ytest = data_type_check(ytest)
        ytest = ytest.reshape(-1).to(self.tkwargs['device'])
        Xtest = Xtest.to(self.tkwargs['device'])
        Xtest, indices = torch.sort(Xtest, dim=0)
        ytest = ytest[indices]

        mean_pred, std_pred = model.predict(Xtest.to(**self.tkwargs), return_std=True)
        confidence_interval = 1.96 * std_pred


        if ax is None:
            fig, ax = plt.subplots()

        ax.scatter(Xtrain, ytrain, s=180, marker='x', color='red', label=Train_label)
        ax.plot(Xtest, ytest, color='black', linewidth=4.0, label=label_test)
        ax.plot(Xtest, mean_pred.cpu(), color='green', linestyle='dashed', linewidth=4.0, label=label_prediction)
        ax.fill_between(Xtest.squeeze(), mean_pred.cpu() - confidence_interval.cpu(), 
                        mean_pred.cpu() + confidence_interval.cpu(), color='lightblue', alpha=0.7, label='95% CI')
        if x_lower is not None and x_upper is not None:
            ax.set_xlim([x_lower, x_upper])
        if y_lower is not None and y_upper is not None:
            ax.set_ylim([y_lower, y_upper])
        ax.set_xlabel(rf'${xlabel}$')
        ax.set_ylabel(rf'${ylabel}$')
        ax.legend()
            
    def evaluation_2(self,Xtest,ytest,n_FP=1):
        Xtest=data_type_check(Xtest)
        ytest=data_type_check(ytest)
        
        ytest=ytest.reshape(-1).to(self.tkwargs['device'])
        Xtest=Xtest.to(self.tkwargs['device'])
        self.eval()
        likelihood=self.likelihood
        likelihood.fidel_indices=self.train_inputs[0][:,-1]
        output=self(Xtest)
        likelihood.fidel_indices=Xtest[:,-1]
        ytest_sc = (ytest-self.y_min)/self.y_std
        mean_temp=[]
        var_temp=[]
        for i in range (n_FP):
            with torch.no_grad():
                trained_pred_dist = likelihood(output)
                mean_temp.append(trained_pred_dist.mean)
                var_temp.append(trained_pred_dist.variance)
            
        sum_list = [mean**2 + var for mean, var in zip(mean_temp, var_temp)]
        sum_tensors = sum(sum_list)/n_FP
        mean_ensamble=sum(mean_temp)/n_FP
        var_ensamble=sum_tensors -mean_ensamble**2
        std_ensamble=var_ensamble.sqrt()
        mu_low, mu_up=mean_ensamble-1.96*std_ensamble, mean_ensamble+1.96*std_ensamble
        final_mse=((ytest_sc.reshape(-1)-mean_ensamble)**2).mean()
        def interval_score(y_true,mu_low, mu_up, alpha = 0.05):
            out = mu_up - mu_low
            out += (y_true > mu_up)* 2/alpha * (y_true - mu_up)
            out += (y_true <mu_low)* 2/alpha * (mu_low - y_true)
            return out
        IS=interval_score(ytest_sc,mu_low, mu_up, alpha = 0.05).mean()
        NIS=IS*torch.abs(self.y_std)/ytest.std()
        NRMSE=torch.sqrt((final_mse*(self.y_std)**2)/ytest.std()**2)    
        table_data = [
        ['NRMSE', NRMSE],
        ['NIS', NIS],
        ]
        # Print the table
        table = tabulate(table_data, headers=['Metric', 'Value'], tablefmt='fancy_grid', colalign=("left", "left"))
        print(table)
        # return NIS, NRMSE


    def rearrange_one_hot(self,tensor):
        # Find the indices that sort each row
        sorted_indices = torch.argsort(tensor, dim=1, descending=True)
        # Generate a new tensor of zeros with the same shape
        new_tensor = torch.zeros_like(tensor)
        # Place '1's in the appropriate positions m_gpd on the sorted indices
        for i in range(tensor.size(0)):
            new_tensor[i, sorted_indices[i, 0]] = 1

        return torch.flip(new_tensor, dims=[0])
    def visualize_latent(self,type='cat',rpearts=500):
        if self.embedding_type=='deterministic':
            if len(self.qual_kernel_columns) > 0:
                for i in range(len(self.qual_kernel_columns)):
                    zeta = self.zeta[i]
                    dimm=zeta.size()[0]
                    zeta_epsilon=torch.normal(mean=0,std=1,size=[dimm,2])

                    A = getattr(self,'A_matrix')
                    positions = A[i](x=zeta.float().to(**self.tkwargs))
                    level = torch.max(self.perm[i]+1, axis = 0)[0].tolist()

                    perm = self.perm[i]
                    plot_sep(type=type,positions = positions, levels = level, perm = perm, constraints_flag=False)
        elif self.embedding_type=='probabilistic':
            for i in range(len(self.qual_kernel_columns)):
                temp= self.transform_categorical(x=self.train_inputs[0][:,self.qual_kernel_columns[i]].clone().type(torch.int64).to(self.tkwargs['device']), perm_dict = self.perm_dict[i], zeta = self.zeta[i])
            unique_rows, indices = torch.unique(temp, dim=0, return_inverse=True)
            xp=self.rearrange_one_hot(unique_rows)
            z_p_list =[]
            label=[]
            epsilon=torch.normal(mean=0,std=1,size=[rpearts,2])
            for i in range(self.num_levels_per_var[0]):
                x_0=xp[i]
                x_0=x_0.repeat(rpearts, 1)
                z_p = getattr(self, 'A_matrix')(x=x_0, epsilon=epsilon)
                z_p_list.append(z_p)
                label.append(i*torch.ones_like(z_p))
            z_p_all = torch.cat(z_p_list, dim=0)
            label_ground_truth=torch.cat(label, dim=0)
            #########################
            plt.rcParams['font.family'] = 'Times New Roman'
            plt.rcParams['font.size'] = 25
            # plt.rcParams['figure.dpi']=150
            tab20 = plt.get_cmap('tab10')
            colors = tab20.colors
            colors=['deeppink','gold','darkorange','gray','orangered']
            plt.figure(figsize=(8,6))

            # Assuming z_p_all is a torch.Tensor
            z_p_all_np = z_p_all.detach().numpy()
            unique_labels = np.unique(label_ground_truth)
            markers = ['X','o','s',"v", 'p']

            for idx, label in enumerate(unique_labels):
                mask = (label_ground_truth == label)
                plt.scatter(z_p_all_np[mask[:,0], 0], z_p_all_np[mask[:,0], 1], 
                            c=colors[idx], 
                            marker=markers[idx], 
                            alpha=.6,
                            s=250, 
                            label=f'Label {label}')

            # Create the legend and get the legend handles and labels
            legend=[ 'HF', 'LF1','LF2','LF3']
            plt.xlabel(r'$z_1$',labelpad=0,rotation=130)
            plt.ylabel(r'$z_2$',labelpad=14,rotation=0)
            plt.tight_layout()
            plt.show()

    
    def evaluation(self,Xtest,ytest):
        Xtest=data_type_check(Xtest)
        ytest=data_type_check(ytest)
        self.eval()
        ytest=ytest.reshape(-1).to(self.tkwargs['device'])
        Xtest=Xtest.to(self.tkwargs['device'])
        ytest=ytest.reshape(-1)
        Xtest=Xtest
        likelihood=self.likelihood
        ytest_sc = (ytest-self.y_min)/self.y_std

        with torch.no_grad():
            trained_pred_dist = likelihood(self(Xtest))
        # Negative Log Predictive Density (NLPD)
        final_nlpd = gpytorch.metrics.negative_log_predictive_density(trained_pred_dist,ytest_sc)
        
        # Mean Squared Error (MSE)
        final_mse = gpytorch.metrics.mean_squared_error(trained_pred_dist, ytest_sc, squared=True)
        # Mean Absolute Error (MAE)
        final_mae = gpytorch.metrics.mean_absolute_error(trained_pred_dist, ytest_sc)
        def interval_score(y_true, trained_pred_dist, alpha = 0.05):
            mu_low, mu_up = trained_pred_dist.confidence_region()
            out = mu_up - mu_low
            out += (y_true > mu_up)* 2/alpha * (y_true - mu_up)
            out += (y_true <mu_low)* 2/alpha * (mu_low - y_true)
            return out
        IS=interval_score(ytest_sc, trained_pred_dist, alpha = 0.05).mean()

        ## back to the original scale:
        final_mse=final_mse*(self.y_std)**2
        final_mae=final_mae*torch.abs(self.y_std)
        IS=IS*torch.abs(self.y_std)
        ###    
        RRMSE=torch.sqrt(final_mse/torch.var(ytest))
        table_data = [
            ['Negative Log-Likelihood (NLL)', final_nlpd],
            ['Mean Squared Error (MSE)', final_mse],
            ['Mean Absolute Error  (MAE)', final_mae],
            ['Relative Root Mean Square Error (RRMSE)', RRMSE],
            ['Interval Score (IS)', IS]
        ]
        # Print the table
        table = tabulate(table_data, headers=['Metric', 'Value'], tablefmt='fancy_grid', colalign=("left", "left"))
        print(table)

    def calibration_result(self,mean_train,std_train):
        self.calibration_id
        n_s=0
        for n in self.calibration_id:
            n_s+=1
            if self.calibration_type=='probabilistic':
                mean= (getattr(self,'Theta_'+str(n)).weights*std_train[n] +mean_train[n])[0].detach().numpy()
                STD= torch.abs((getattr(self,'Theta_'+str(n)).bias)*std_train[n])[0].detach().numpy()
                print("For Calibration parameter Theta_"+str(n)+ " Estimated Mean is "  + str(mean)+" and Estimated STD is "  + str(STD))
                x = np.linspace(mean-5*STD,mean+5*STD, 1000)
                pdf_values = norm.pdf(x, mean, STD).squeeze()
                plt.figure()
                plt.plot(x, pdf_values, label='Zeta_'+str(n))
                plt.title(r'$\mathit{\hat{\Theta}}_{' + str(n) + '}$')  # Italic LaTeX styled title with hat only on Theta
                plt.xlabel('Value')
                plt.ylabel('Density')
                plt.grid(True) 

            else:
                xx=torch.where(self.train_inputs[0][:,-1]==0)[0][0]
                GT=self.train_inputs[0][xx,n]*std_train[n] +mean_train[n]
                print("=================GP + Results===================")
                print("Estimated Calibration parameter for Zeta_"+str(n_s)+ " is "  + str((getattr(self,'Theta_'+str(n)).constant.detach()*std_train[n] +mean_train[n])))

    @classmethod
    def show(cls):
        plt.show()
        
    def get_params(self, name = None):
        params = {}
        print('###################Parameters###########################')
        for n, value in self.named_parameters():
             params[n] = value
        if name is None:
            print(params)
            return params
        else:
            if name == 'Mean':
                key = 'mean_module.constant'
            elif name == 'Sigma':
                key = 'covar_module.raw_outputscale'
            elif name == 'Noise':
                key = 'likelihood.noise_covar.raw_noise'
            elif name == 'Omega':
                for n in params.keys():
                    if 'raw_lengthscale' in n and params[n].numel() > 1:
                        key = n
            print(params[key])
            return params[key]
    

    def sample_y(self, size = 1, X = None, plot = False):
        if X == None:
            X = self.train_inputs[0]
        
        self.eval()
        out = self.likelihood(self(X))
        draws = out.sample(sample_shape = torch.Size([size]))
        index = np.argsort(out.loc.detach().numpy())
        if plot:
            _ = plt.figure(figsize=(12,6))
            _ = plt.scatter(list(range(len(X))), out.loc.detach().numpy()[index], color = 'red', s = 20, marker = 'o')
            _ = plt.scatter(np.repeat(np.arange(len(X)).reshape(1,-1), size, axis = 0), 
                draws.detach().numpy()[:,index], color = 'blue', s = 1, alpha = 0.5, marker = '.')
        return draws

    def get_latent_space(self):
        positions=[]
        if len(self.qual_dict_list) > 0:
            for i in range(len(self.qual_kernel_columns)):
                zeta = self.zeta[i]
                A = getattr(self,'A_matrix')
                positions.append(A[i](x=zeta.float().to(**self.tkwargs)).detach())
            return positions
        else:
            print('No categorical Variable, No latent positions')
            return None

        



    def LMMAPPING(self, num_features:int, type = 'Linear',embedding_dim = 2):

        if type == 'Linear':
            in_feature = num_features
            out_feature = embedding_dim
            lm = torch.nn.Linear(in_feature, out_feature, bias = False)
            return lm

        else:
            raise ValueError('Only Linear type for now')    

    def zeta_matrix(self,
        num_levels:int,
        embedding_dim:int,
        batch_shape=torch.Size()
    ) -> None:

        if any([i == 1 for i in num_levels]):
            raise ValueError('Categorical variable has only one level!')

        if embedding_dim == 1:
            raise RuntimeWarning('1D latent variables are difficult to optimize!')
        
        for level in num_levels:
            if embedding_dim > level - 0:
                embedding_dim = min(embedding_dim, level-1)
                raise RuntimeWarning(
                    'The LV dimension can atmost be num_levels-1. '
                    'Setting it to %s in place of %s' %(level-1,embedding_dim)
                )
    
        from itertools import product
        levels = []
        for l in num_levels:
            levels.append(torch.arange(l))

        perm = list(product(*levels))
        perm = torch.tensor(perm, dtype=torch.int64)

        #-------------Mapping-------------------------
        perm_dic = {}
        for i, row in enumerate(perm):
            temp = str(row.tolist())
            if temp not in perm_dic.keys():
                perm_dic[temp] = i

        #-------------One_hot_encoding------------------
        for ii in range(perm.shape[-1]):
            if perm[...,ii].min() != 0:
                perm[...,ii] -= perm[...,ii].min()
            
        perm_one_hot = []
        for i in range(perm.size()[1]):
            perm_one_hot.append( torch.nn.functional.one_hot(perm[:,i]) )

        perm_one_hot = torch.concat(perm_one_hot, axis=1)

        return perm_one_hot, perm, perm_dic

    #################################### transformation functions####################################

    def transform_categorical(self, x:torch.Tensor,perm_dict = [], zeta = []) -> None:
        if x.dim() == 1:
            x = x.reshape(-1,1)
        # categorical should start from 0
        if self.training == False:
            x = torch.tensor(setlevels(x))
        if self.encoding_type == 'one-hot':
            try:
                index = [perm_dict[str(row.tolist())] for row in x]
            except KeyError:
                raise ValueError(
                    "The categorical input (or source indices) are not defined properly. "
                    "They should be integer values starting from zero. To solve the issue, "
                    "you can use the 'setlevels' function, which is a preprocessing function."
                )
            if x.dim() == 1:
                x = x.reshape(len(x),)

            return zeta[index,:]  

    def transform_categorical_random_varible_for_latent(self,x_raw, x:torch.Tensor,perm_dict = [], zeta = []) -> None:
        
        dimm=zeta.size()[0]
        # zeta=torch.normal(0,1,size=[dimm,2])

        self.random_zeta.append(torch.normal(mean=0,std=1,size=[dimm,2]))
        
        if x_raw.requires_grad:
            random_zeta_appy=self.random_zeta[-1]
        else:
            random_zeta_appy=self.random_zeta[0]
        
        if x.dim() == 1:
            x = x.reshape(-1,1)
        # categorical should start from 0
        if self.training == False:
            x = setlevels(x)
        if self.encoding_type == 'one-hot':
            index = [perm_dict[str(row.tolist())] for row in x]

            if x.dim() == 1:
                x = x.reshape(len(x),)

            return random_zeta_appy[index,:]  
        
    def final_transform_categorical_random_varible_for_latent(self,x_raw, x:torch.Tensor,perm_dict = [], zeta = []) -> None:
        
        if x.dim() == 1:
            x = x.reshape(-1,1)
        if self.training == False:
            x = setlevels(x)
        if self.encoding_type == 'one-hot':
            index = [perm_dict[str(row.tolist())] for row in x]

            if x.dim() == 1:
                x = x.reshape(len(x),)

        dimm=zeta.size()[0]

        self.random_zeta.append(torch.normal(mean=0,std=1,size=[dimm,2]))
        
        if x_raw.requires_grad:
            random_zeta_appy=self.random_zeta[-1]
        else:
            if x_raw.size()[0]==400:
                random_zeta_appy=0*self.random_zeta[0]
            else: 
                random_zeta_appy=self.random_zeta[0]
        return random_zeta_appy[index,:]       
    

    def Sobol(self, N=10000):
        """

        This function calculates the sensitivity indecies for a function
        Inputs:
            self (GP_Model): The GP model (fitted by GP+) with p inputs and dy outputs.

            N: is the size of the Sobol sequence used for evaluating the indecies. Should be larger than 1e5 for accuracy.

        Outputs:
            S: Matrix of size dy-by-p of main sensitivity indecies. 
            ST: Matrix of size dy-by-p of total sensitivity indecies.
        """
        if N<1e5:
            warnings.warn('Increase N for accuracy!')

        p = self.train_inputs[0].shape[1] 
        dy = 1# self.train_targets.shape[1] 

        self.qual_dict_list
        self.num_levels_per_var
        # sequence = torch.from_numpy( sobol_seq.i4_sobol_generate(2*p, N)).to(**self.tkwargs)
        sequence = torch.from_numpy( sobol_seq.i4_sobol_generate(2*p, N))
        def normalize_sobol_sequence(sequence, train_inputs,p):
            
            temp_1 = sequence[:,p:]
            temp_2 = sequence[:,:p]
            
            # Normalize the sequence
            mins = train_inputs.min(dim=0)[0]
            maxs = train_inputs.max(dim=0)[0]

            sequence_1= mins + (maxs - mins) * temp_1
            sequence_2= mins + (maxs - mins) * temp_2
            # Take care of categotrical inputes
            j=0
            for i in self.qual_dict_list:
                temp_1[:,i]= temp_1[:,i]*(self.num_levels_per_var[j]-1)
                temp_2[:,i]=temp_2[:,i]*(self.num_levels_per_var[j]-1)
                sequence_1[:,i]=temp_1[:,i].round()
                sequence_2[:,i]=temp_2[:,i].round()
                j+=1
                return sequence_1,sequence_2
        A,B = normalize_sobol_sequence(sequence, self.train_inputs[0],p)

        # # A = A * (self.Y.max(axis=0) - self.Y.min(axis=0)) + self.Y.min(axis=0) ## Normalize genrated data

        # B = A[:,p:]
        # A = A[:,:p]
        
        AB = torch.zeros((N,p,p))
        for i in range(p):
            AB[:,:,i] = A
            AB[:,i,i] = B[:,i]
            
        FA = self.predict(A,return_std=False).detach().cpu().numpy().reshape(-1,1)

        FB = self.predict(B,return_std=False).detach().cpu().numpy().reshape(-1,1)

        FAB = np.zeros((N, p, dy))
        for i in range(p):
            temp = self.predict(AB[:, :, i],return_std=False).detach().cpu().numpy()
            FAB[:, i, :] = temp.reshape(-1,1)

        S = np.zeros((p, dy))
        ST = np.zeros((p, dy))

        for i in range(p):
            temp = FAB[:, i, :]
            S[i, :] = np.sum(FB * (temp - FA), axis=0) / N
            ST[i, :] = np.sum((FA - temp)**2, axis=0) / (2 * N)
            
        varY = np.var(np.concatenate([FA,FB]), axis=0)
        S = (S / varY).T
        ST = (ST / varY).T

        return S, ST

######################################################################## Other Classes Used in GP_Pluse  #####################################################
class FFNN(nn.Module):
    def __init__(self, GP_Plus, input_size, num_classes, layers,name):
        super(FFNN, self).__init__()
        self.hidden_num = len(layers)
        if self.hidden_num > 0:
            self.fci = nn.Linear(input_size, layers[0], bias=False) 
            GP_Plus.register_parameter(str(name)+'fci', self.fci.weight)
            GP_Plus.register_prior(name = 'latent_prior_fci', prior=gpytorch.priors.NormalPrior(0.,1), param_or_closure=str(name)+'fci')

            for i in range(1,self.hidden_num):
                setattr(self, 'h' + str(i), nn.Linear(layers[i-1], layers[i], bias=False))
                GP_Plus.register_parameter(str(name)+'h'+str(i), getattr(self, 'h' + str(i)).weight )
                GP_Plus.register_prior(name = 'latent_prior'+str(i), prior=gpytorch.priors.NormalPrior(0.,1), param_or_closure=str(name)+'h'+str(i))
            
            self.fce = nn.Linear(layers[-1], num_classes, bias= False)
            GP_Plus.register_parameter(str(name)+'fce', self.fce.weight)
            GP_Plus.register_prior(name = 'latent_prior_fce', prior=gpytorch.priors.NormalPrior(0.,1), param_or_closure=str(name)+'fce')
        else:
            self.fci = Linear_MAP(input_size, num_classes, bias = False)
            GP_Plus.register_parameter(name, self.fci.weight)
            GP_Plus.register_prior(name = 'latent_prior_'+name, prior=gpytorch.priors.NormalPrior(0,1) , param_or_closure=name)

    def forward(self, x, transform = lambda x: x):
        """
        x here is the mnist images and we run it through fc1, fc2 that we created above.
        we also add a ReLU activation function in between and for that (since it has no parameters)
        I recommend using nn.functional (F)
        """
        if self.hidden_num > 0:
            x = torch.tanh(self.fci(x))
            for i in range(1,self.hidden_num):
                #x = F.relu(self.h(x))
                x = torch.tanh( getattr(self, 'h' + str(i))(x) )
            
            x = self.fce(x)
        else:
            #self.fci.weight.data = torch.sinh(self.fci.weight.data)
            x = self.fci(x, transform)
        return x
    
############################################
class FFNN_as_Mean(gpytorch.Module):
    def __init__(self, GP_Plus, input_size, num_classes, layers,name):
        super(FFNN_as_Mean, self).__init__()
        self.dropout = nn.Dropout(0.2)
        self.hidden_num = len(layers)
        if self.hidden_num > 0:
            self.fci = Linear_class(input_size, layers[0], bias=True, name='fci') 
            for i in range(1,self.hidden_num):
                setattr(self, 'h' + str(i), Linear_class(layers[i-1], layers[i], bias=True,name='h' + str(i)))
            
            self.fce = Linear_class(layers[-1], num_classes, bias=True,name='fce')
        else:
            self.fci = Linear_class(input_size, num_classes, bias=True, dtype = torch.float32,name='fci') #Linear_MAP(input_size, num_classes, bias = True)

    def forward(self, x, transform = lambda x: x):

        if self.hidden_num > 0:
            
            x = torch.tanh(self.fci(x))
            # x = self.dropout(x)
            # x = self.fci(x)
            for i in range(1,self.hidden_num):
                # x = torch.sigmoid( getattr(self, 'h' + str(i))(x) )
                # x =  getattr(self, 'h' + str(i))(x) 
                x = torch.tanh( getattr(self, 'h' + str(i))(x) )
                x = self.dropout(x)
            x = self.fce(x)
        else:
            #self.fci.weight.data = torch.sinh(self.fci.weight.data)
            x = self.fci(x)

        return x
    
############################################
class Linear_VAE(Mean):
    __constants__ = ['in_features', 'out_features']
    in_features: int
    out_features: int
    weight: Tensor
    def __init__(self, in_features: int, out_features: int, bias: bool = True, name=None,
                 device=None, dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super(Linear_VAE, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.name=str(name)
        self.register_parameter(name=str(self.name)+'weight',  parameter= Parameter(torch.empty((out_features, in_features), **factory_kwargs)))
        self.register_prior(name =str(self.name)+ 'prior_m_weight_fci', prior=gpytorch.priors.NormalPrior(0.,.2), param_or_closure=str(self.name)+'weight')

        if bias:

            self.register_parameter(name=str(self.name)+'bias',  parameter=Parameter(torch.empty(out_features, **factory_kwargs)))
            self.register_prior(name= str(self.name)+'prior_m_bias_fci', prior=gpytorch.priors.NormalPrior(0.,.05), param_or_closure=str(self.name)+'bias')
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self) -> None:                                             

        init.kaiming_uniform_( getattr(self,str(self.name)+'weight'), a=math.sqrt(5))
        if getattr(self,str(self.name)+'bias') is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(getattr(self,str(self.name)+'weight'))
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(getattr(self,str(self.name)+'bias'), -bound, bound)

    def forward(self, input) -> Tensor:

        return F.linear(input.double(), getattr(self,str(self.name)+'weight').double(), getattr(self,str(self.name)+'bias').double())      ### Forced to Add .double() for NN in mean function

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        )

############################################
class LinearVariational(Mean):
    def __init__(self, batch_shape=torch.Size(),mean_prior=1,std_prior=1):
        super().__init__()
        self.register_parameter(name="weights", parameter=torch.nn.Parameter(torch.randn(*batch_shape,1,1)))
        self.register_prior(name = 'weights_prior', prior=gpytorch.priors.NormalPrior(mean_prior,1), param_or_closure='weights')
        self.register_parameter(name="bias", parameter=torch.nn.Parameter(torch.randn(*batch_shape, 1,1)))
        self.register_prior(name = 'bias_prior', prior=gpytorch.priors.NormalPrior(std_prior,1), param_or_closure='bias')

    def forward(self, epsilon):
        res = self.weights + (torch.abs(self.bias)) *epsilon
        return res

##########################################
class LinearMean_with_prior(Mean):
    def __init__(self, input_size, batch_shape=torch.Size(), bias=True):
        super().__init__()
        self.register_parameter(name="weights", parameter=torch.nn.Parameter(torch.randn(*batch_shape, input_size, 1)))
        self.register_prior(name = 'weights_prior', prior=gpytorch.priors.NormalPrior(0.,.5), param_or_closure='weights')
        if bias:
            self.register_parameter(name="bias", parameter=torch.nn.Parameter(torch.randn(*batch_shape, 1)))
            self.register_prior(name = 'bias_prior', prior=gpytorch.priors.NormalPrior(0.,.5), param_or_closure='bias')
        else:
            self.bias = None
    def forward(self, x):
        res = x.matmul(self.weights).squeeze(-1)
        if self.bias is not None:
            res = res + self.bias
        return res
    
############################################
class Variational_Encoder(gpytorch.Module):
    def __init__(self, GP_Plus, input_size, num_classes, layers,name):
        super(Variational_Encoder, self).__init__() 
        self.dropout = nn.Dropout(0.2)
        self.hidden_num = len(layers)
        if self.hidden_num > 0:
            self.fci = Linear_VAE(input_size, layers[0], bias=True, name='fci') 
            for i in range(1,self.hidden_num):
                #self.h = nn.Linear(neuran[i-1], neuran[i])
                setattr(self, 'h' + str(i), Linear_VAE(layers[i-1], layers[i], bias=True,name='h' + str(i)))
            self.fce = Linear_VAE(layers[-1], num_classes, bias=True,name='fce')
        else:
            self.fci = Linear_VAE(input_size, num_classes, bias=True, dtype = torch.float32,name='fci') 

    def forward(self, x,epsilon, transform = lambda x: x):
        if self.hidden_num > 0:
            # x = torch.tanh(self.fci(x))
            x =self.fci(x)
            for i in range(1,self.hidden_num):
                # x = F.relu(self.h(x))
                x = torch.tanh( getattr(self, 'h' + str(i))(x) )
                # x = self.dropout(x)
            output = self.fce(x)

            epsilon_1, epsilon_2 = epsilon[:, 0:1], epsilon[:, 1:2]
            L22, L21, L11, Mu_2, Mu_1 = output[:, 0:1], output[:, 1:2], output[:, 2:3], output[:, 3:4], output[:, 4:5]
            # Optimized calculation using matrix operations
            X_1 = Mu_1 + 1*torch.abs(L11) * epsilon_1
            X_2 = Mu_2 + 1*L21 * epsilon_1 + 1*torch.abs(L22) * epsilon_2
            x = torch.cat((X_1,X_2),1)

        else:  
            output = self.fci(x)
            epsilon_1, epsilon_2 = epsilon[:, 0:1], epsilon[:, 1:2]
            L22, L21, L11, Mu_2, Mu_1 = output[:, 0:1], output[:, 1:2], output[:, 2:3], output[:, 3:4], output[:, 4:5]
            # calculation using matrix operations
            X_1 = Mu_1 + 1*torch.abs(L11) * epsilon_1
            X_2 = Mu_2 + 1*L21 * epsilon_1 + 1*torch.abs(L22) * epsilon_2
            x = torch.cat((X_1,X_2),1)
        return x 
    
############################################
class Linear_class(Mean):
    __constants__ = ['in_features', 'out_features']
    in_features: int
    out_features: int
    weight: Tensor

    def __init__(self, in_features: int, out_features: int, bias: bool = True, name=None,
                 device=None, dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super(Linear_class, self).__init__()
        self.in_features = in_features
        self.out_features = out_features

        self.name=str(name)
        self.register_parameter(name=str(self.name)+'weight',  parameter= Parameter(torch.empty((out_features, in_features), **factory_kwargs)))
        self.register_prior(name =str(self.name)+ 'prior_m_weight_fci', prior=gpytorch.priors.NormalPrior(0.,0.01), param_or_closure=str(self.name)+'weight')
        if bias:
            self.register_parameter(name=str(self.name)+'bias',  parameter=Parameter(torch.empty(out_features, **factory_kwargs)))
            self.register_prior(name= str(self.name)+'prior_m_bias_fci', prior=gpytorch.priors.NormalPrior(0.,.001), param_or_closure=str(self.name)+'bias')
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self) -> None:                                         
        init.kaiming_uniform_( getattr(self,str(self.name)+'weight'), a=math.sqrt(5))
        if getattr(self,str(self.name)+'bias') is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(getattr(self,str(self.name)+'weight'))
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(getattr(self,str(self.name)+'bias'), -bound, bound)

    def forward(self, input) -> Tensor:

        # return F.linear(input, getattr(self,str(self.name)+'weight').double(), getattr(self,str(self.name)+'bias').double())      ### Forced to Add .double() for NN in mean function
        return F.linear(input, getattr(self,str(self.name)+'weight'), getattr(self,str(self.name)+'bias'))      ### Forced to Add .double() for NN in mean function

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        )
    
############################################
class Linear_MAP(nn.Linear):
    def __init__(self, in_features: int, out_features: int, bias: bool = True, device=None, dtype=None) -> None:
        super().__init__(in_features, out_features, bias, device, dtype)
        
    def forward(self, input, transform = lambda x: x):
        return F.linear(input,transform(self.weight), self.bias)



# LMGP Visualization
# 
#
#

from turtle import color
import numpy as np
import matplotlib.pyplot as plt
import torch


tkwargs = {
    "dtype": torch.double,
    "device": torch.device("cpu" if torch.cuda.is_available() else "cpu"),
}


def plot_ls(model, constraints_flag = True):
    # 
    # plot latent values

    zeta = torch.tensor(model.zeta, dtype = torch.float64).to(**tkwargs)
    #A = model.nn_model.weight.detach()
    perm = model.perm
    levels = model.num_levels_per_var
    #positions = torch.matmul(zeta, A.T)   # this gives the position of each combination in latent space

    positions = model.nn_model(zeta)
    positions = positions.detach()

    # applying the constrains
    if constraints_flag:
        positions = constrains(positions.to(**tkwargs))


    fig,axs = plt.subplots(1, len(levels),figsize=(12,6))
    #colors = {0:'blue', 1:'r', 2:'g', 3:'c', 4:'m', 5:'k', 6:'y'}
    tab20 = plt.get_cmap('tab10')
    colors = tab20.colors

    # loop over the number of variables
    for j in range(len(levels)):

        for i in range(levels[j]):
            index = torch.where(perm[:,j] == i) 
            #col = list(map(lambda x: colors[x], np.ones(index[0].numpy().shape) * i))
            axs[j].scatter(positions.cpu().numpy()[index][:,0], positions.cpu().numpy()[index][:,1], label = 'level' + str(i+1), color = colors[i%10], s = (i+1)*50/levels[j])
            axs[j].set_title('Variable ' + str(j), fontsize = 15)
            axs[j].set_xlabel(r'$z_1$', fontsize = 15)
            axs[j].set_ylabel(r'$z_2$', fontsize = 15)
            axs[j].legend()

        fig.tight_layout()


def constrains(z):
    n = z.shape[0]
    z = z - z[0,:]

    if z[1,0] < 0:
        z[:, 0] *= -1
    
    rot = torch.atan(-1 * z[1,1]/z[1,0])
    R = torch.tensor([ [torch.cos(rot), -1 * torch.sin(rot)], [torch.sin(rot), torch.cos(rot)]]).to(**tkwargs)

    z = torch.matmul(R, z.T)
    z = z.T
    if z.shape[1] > 2 and z[2,1] < 0:
        z[:, 1] *= -1
    
    return z


import torch
import numpy as np
def data_type_check(data):
    if isinstance(data, torch.Tensor):
        return data
    elif isinstance(data, np.ndarray):
        print(f"Warning: Data type was numpy.ndarray. GP+ made it a torch tensor to be able to continue.")
        return torch.from_numpy(data)
    else:
        print(f"Warning: Data type was {type(data)}. GP+ made it a torch tensor to be able to continue.")
        return torch.tensor(data)


# Copyright © 2023, Dr. Bostanabad's research group at the University of California, Irvine.
# 
# GP+ Intellectual Property Notice:
# 
# The software known as GP+ is the proprietary material of Dr. Bostanabad's research group at the University of California, Irvine. 
# Non-profit academic institutions and U.S. government agencies may utilize this software exclusively for educational and research endeavors. 
# All other entities are granted permission for evaluation purposes solely; any additional utilization demands prior written consent from the appropriate authority. 
# The direct sale or redistribution of this software, in any form, without explicit written authorization is strictly prohibited. 
# Users are permitted to make duplicate copies of the software, contingent upon the assurance that no copies are sold or redistributed and they adhere to the stipulated terms herein.
# 
# Being academic research software, GP+ is provided on an "as is" m_gp, devoid of warranties, whether explicit or implicit. 
# The act of downloading or executing any segment of this software inherently signifies compliance with these terms. 
# The developers reserve the right to modify these terms and conditions without prior intimation at any juncture.

import torch
import math
from gpytorch.mlls import ExactMarginalLogLikelihood
from typing import Dict,List,Optional
from copy import deepcopy
from tqdm import tqdm
import numpy as np

def get_bounds(likobj, theta):

    dic = likobj.unpack_parameters(theta)

    minn = np.empty(0)
    maxx = np.empty(0)
    for name, values in dic.items():
        # print(name)
        for ii in range(len(likobj.model.qual_kernel_columns)):
            if name ==  str(likobj.model.qual_kernel_columns[ii]):
                minn = np.concatenate( (minn,  np.repeat(-3, values.numel()) ) )
                maxx = np.concatenate( (maxx,  np.repeat(3, values.numel()) ) )
        if name == 'likelihood.noise_covar.raw_noise':
            minn = np.concatenate( (minn,  np.repeat(-np.inf, values.numel()) ) )
            maxx = np.concatenate( (maxx,  np.repeat( np.inf, values.numel()) ) )
        elif name == 'covar_module.raw_outputscale':
            minn = np.concatenate( (minn,  np.repeat(0, values.numel()) ) )
            maxx = np.concatenate( (maxx,  np.repeat( np.inf, values.numel()) ) )
        elif 'raw_lengthscale' in name:
            minn = np.concatenate( (minn,  np.repeat(-10.0, values.numel()) ) )
            maxx = np.concatenate( (maxx,  np.repeat( 3.0, values.numel()) ) )
        elif name == 'covar_module.base_kernel.kernels.2.raw_lengthscale':
            maxx = np.concatenate( (maxx,  np.repeat( 3.0, values.numel()) ) )
        elif name == 'A_matrix.fci.[10]fci':
            minn = np.concatenate( (minn,  np.repeat(-np.inf, values.numel()) ) )
            maxx = np.concatenate( (maxx,  np.repeat( np.inf, values.numel()) ) )
            ######################################################################################### For multiple Bases ##################################
        elif name[0:4] == 'mean':
            minn = np.concatenate( (minn,  np.repeat(-1.5, values.numel()) ) )
            maxx = np.concatenate( (maxx,  np.repeat( 1.5, values.numel()) ) )
    return np.array(minn).reshape(-1,), np.array(maxx).reshape(-1,)


def fit_model_torch(
    model,
    model_param_groups:Optional[List]=None,
    lr_default:float=0.01,
    num_iter:int=100,
    num_restarts:int=0,
    break_steps:int = 50) -> float:
    '''Optimize the likelihood/posterior of a standard GP+ model using `torch.optim.Adam`.

    This is a convenience function that covers many situations for optimizing a standard GP model.
    Note that using L-BFGS through `fit_model_scipy` function is a better optimization strategy.

    :param model: A model instance derived from the `models.GPR` class. Can also pass a instance
        inherting from `gpytorch.models.ExactGP` provided that `num_restarts=0` or 
        the class implements a `.reset_parameters` method.
    :type model: models.GPR

    :param model_param_groups: list of parameters to optimizes or dicts defining parameter
        groups. If `None` is specified, then all parameters with `.requires_grad`=`True` are 
        included. Defaults to `None`.
    :type model_param_groups: list, optional

    :param lr_default: The default learning rate for all parameter groups. To use different 
        learning rates for some groups, specify them `model_param_groups`. 
    :type lr_default: float, optional

    :param num_iter: The number of optimization steps from each starting point. This is the only
        termination criterion for the optimizer.
    :type num_iter: float, optional

    :param num_restarts: The number of times to restart the local optimization from a 
        new starting point. Defaults to 5
    :type num_restarts: int, optional

    :returns: the best (negative) log-likelihood/log-posterior found
    :rtype: float
    '''  
    model.train()
    
    # objective
    mll = ExactMarginalLogLikelihood(model.likelihood, model)
    f_inc = math.inf
    current_state_dict = model.state_dict()


    loss_hist_total = []
    # ############################  ADAM ################################

    for i in range(num_restarts+1):
        optimizer = torch.optim.Adam(
            model.parameters() if model_param_groups is None else model_param_groups, 
            lr=lr_default)
        loss_hist = []
        epochs_iter = tqdm(range(num_iter),desc='Epoch',position=0,leave=True)
        for j in epochs_iter:
            # zero gradients from previous iteration
            optimizer.zero_grad()
            # output from model
            output = model(*model.train_inputs)
            # calculate loss and backprop gradients
            loss = -mll(output,model.train_targets)
            loss.backward()
            optimizer.step()

            acc_loss = loss.item()
            desc = f'Epoch {j} - loss {acc_loss:.4f}'
            epochs_iter.set_description(desc)
            epochs_iter.update(1)
            loss_hist.append(acc_loss)

            if j > break_steps and j%break_steps == 0:
                if ( (torch.mean(torch.Tensor(loss_hist)[j-break_steps:j]) - loss_hist[j]) <= 0 ):
                    break
        
        loss_hist_total.append(loss_hist)

        if loss.item()<f_inc:
            current_state_dict = deepcopy(model.state_dict())
            f_inc = loss.item()
        
        if i < num_restarts:
            model.reset_parameters()
    
    model.load_state_dict(current_state_dict)

    return f_inc, loss_hist_total




import numpy as np
from scipy.stats.qmc import Sobol, scale
from gpplus.preprocessing import setlevels

####################################Wing Function################################################
def wing(n=100, X = None, noise_std = 0.0, random_state = None, shuffle = True):

    if random_state is not None:
        np.random.seed(random_state)

    dx = 10
    l_bound = [150, 220, 6, -10, 16, 0.5, 0.08, 2.5, 1700, 0.025]
    u_bound = [200, 300, 10, 10, 45, 1, 0.18, 6, 2500, 0.08]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)
    Sw = X[..., 0]
    Wfw = X[..., 1]
    A = X[..., 2]
    Gama = X[..., 3] * (np.pi/180.0)
    q = X[..., 4]
    lamb = X[..., 5]
    tc = X[..., 6]
    Nz = X[..., 7]
    Wdg = X[..., 8]
    Wp = X[..., 9]
    # This is the output
    y = 0.036 * Sw**0.758 * Wfw**0.0035 * (A/(np.cos(Gama)) ** 2) ** 0.6 * \
        q**0.006*lamb**0.04 * ((100 * tc)/(np.cos(Gama)))**(-0.3) *\
        (Nz * Wdg) ** 0.49 + Sw * Wp


    if shuffle:
        index = np.random.randint(0, len(y), size = len(y))
        X = X[index,...]
        y = y[index]


    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y


####################################Borehole Function################################################
def borehole(n=100, X = None, noise_std = 0.0, random_state = None, shuffle = True):

    if random_state is not None:
        np.random.seed(random_state)

    dx = 8
    l_bound = [0.05, 100, 63070, 990, 63.1, 700, 1120, 9855]
    u_bound = [0.15, 50000, 115600, 1110, 116, 820, 1680, 12045]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)
    rw = X[..., 0]
    r = X[..., 1]
    Tu = X[..., 2]
    Hu = X[..., 3] 
    Tl = X[..., 4]
    Hl = X[..., 5]
    L = X[..., 6]
    Kw = X[..., 7]

    frac1 = 2 * np.pi * Tu * (Hu-Hl)

    frac2a = 2*L*Tu / (np.log(r/rw)*rw**2*Kw)
    frac2b = Tu / Tl
    frac2 = np.log(r/rw) * (1+frac2a+frac2b)

    y = frac1 / frac2


    if shuffle:
        index = np.random.randint(0, len(y), size = len(y))
        X = X[index,...]
        y = y[index]

    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y

####################################Borehole Function################################################

def borehole_mixed_variables(n=100, X = None, qual_dict = {0:5, 6:3}, 
    noise_std = 0.0, random_state = None, shuffle = True):

    labels = {'rw':0, 'r':1, 'Tu':2, 'Hu':3, "Tl":4, 'Hl':5, 'L':6, 'Kw':7}

    dx = 8
    l_bound = [0.05, 100, 63070, 990, 63.1, 700, 1120, 9855]
    u_bound = [0.15, 50000, 115600, 1110, 116, 820, 1680, 12045]
    out_flag = 0
    data = {}
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        # for categorical variables we select t1 levels from the boundary
        for key, value in qual_dict.items():
            levels = np.random.uniform(l_bound[key], u_bound[key], size = value)
            X[...,key] = np.random.choice(levels, size = len(X), replace=True)

        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)
    data['rw'] = X[..., 0]
    data['r'] = X[..., 1]
    data['Tu'] = X[..., 2]
    data['Hu'] = X[..., 3] 
    data['Tl'] = X[..., 4]
    data['Hl'] = X[..., 5]
    data['L'] = X[..., 6]
    data['Kw'] = X[..., 7]

    frac1 = 2 * np.pi * data['Tu'] * (data['Hu']-data['Hl'])

    frac2a = 2*data['L']*data['Tu'] / (np.log(data['r']/data['rw'])*data['rw']**2*data['Kw'])
    frac2b = data['Tu'] / data['Tl']
    frac2 = np.log(data['r']/data['rw']) * (1+frac2a+frac2b)

    y = frac1 / frac2

    if shuffle:
        index = np.random.randint(0, len(y), size = len(y))
        X = X[index,...]
        y = y[index]

    # This will assign levels to categorical evenif the levels are strings
    X = setlevels(X, qual_index = list(qual_dict.keys()))

    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y


####################################### Wing_mixed #######################################
def wing_mixed_variables(n=100, X = None, qual_dict = {0:5, 6:3}, noise_std = 0.0, random_state = None, shuffle = True):

    if random_state is not None:
        np.random.seed(random_state)

    labels = {'Sw':0, 'Wfw':1, 'A':2, 'Gama':3, "q":4, 'lamb':5, 'tc':6, 'Nz':7, 'Wdg':8, 'Wp':9}

    dx = 10
    l_bound = [150, 220, 6, -10, 16, 0.5, 0.08, 2.5, 1700, 0.025]
    u_bound = [200, 300, 10, 10, 45, 1, 0.18, 6, 2500, 0.08]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        for key, value in qual_dict.items():
            levels = np.random.uniform(l_bound[key], u_bound[key], size = value)
            X[...,key] = np.random.choice(levels, size = len(X), replace=True)
            
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)
    Sw = X[..., 0]
    Wfw = X[..., 1]
    A = X[..., 2]
    Gama = X[..., 3] * (np.pi/180.0)
    q = X[..., 4]
    lamb = X[..., 5]
    tc = X[..., 6]
    Nz = X[..., 7]
    Wdg = X[..., 8]
    Wp = X[..., 9]
    # This is the output
    y = 0.036 * Sw**0.758 * Wfw**0.0035 * (A/(np.cos(Gama)) ** 2) ** 0.6 * \
        q**0.006*lamb**0.04 * ((100 * tc)/(np.cos(Gama)))**(-0.3) *\
        (Nz * Wdg) ** 0.49 + Sw * Wp

    if shuffle:
        index = np.random.randint(0, len(y), size = len(y))
        X = X[index,...]
        y = y[index]

    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y


#################################### Sine Function ################################################
def sine_1D(n=100, X = None, noise_std = 0.0, frequency = 1.0, absolute_value_flag = False, random_state = None, shuffle = True):
    if random_state is not None:
        np.random.seed(random_state)
    dx = 1
    l_bound = [-1.0]
    u_bound = [1.0]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)
    
    y = np.sin(2*np.pi*frequency*X[:,0])

    if absolute_value_flag == True:
        y = np.abs(y)

    if shuffle:
        index = np.random.randint(0, len(y), size = len(y))
        X = X[index,...]
        y = y[index]

    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y



# LMGP Visualization
# 
#
#

import numpy as np
import matplotlib.pyplot as plt
import torch

def plot_ls(model, constraints_flag = True):
    # 
    # plot latent values

    #zeta = torch.tensor(model.zeta, dtype = torch.float64)
    zeta = model.zeta

    #A = model.nn_model.weight.detach()
    perm = model.perm
    levels = model.num_levels_per_var
    #positions = torch.matmul(zeta, A.T)   # this gives the position of each combination in latent space

    positions = model.nn_model(zeta, transform = lambda x: x) #3-torch.exp(x)
    if positions.ndim > 2:
        positions = positions.mean(axis = 0)
    else:
        positions = positions.detach()

    if positions.ndim > 2:
        positions = positions.mean(axis = 0) 

    # applying the constrains
    if constraints_flag:
        positions = constrains(positions)


    positions = positions.detach().numpy()

    # plt.rcParams.update({'font.size': 19})
    # fig,axs = plt.subplots(figsize=(8.5,6))
    # colors = {0:'blue', 1:'r', 2:'g', 3:'c', 4:'m', 5:'k', 6:'y'}
    # # loop over the number of variables
    # for j in range(len(levels)):

    #     for i in range(levels[j]):
    #         index = torch.where(perm[:,j] == i) 
    #         col = list(map(lambda x: colors[x], np.ones(index[0].numpy().shape) * i))
    #         axs.scatter(positions[index][...,0], positions[index][...,1], label = 'level' + str(i+1), c = col)
    #         #axs.set_title('Variable ' + str(j), fontsize = 15)
    #         axs.set_xlabel(r'$z_1$', fontsize = 25)
    #         axs.set_ylabel(r'$z_2$', fontsize = 25)
    #         axs.legend()
    #         tempxi = np.min(positions[...,0])-0.2 * (np.abs(np.min(positions[...,0])) +5)
    #         tempxx = np.max(positions[...,0]) + 0.2 * (np.abs(np.max(positions[...,0])) +5)
    #         tempyi = np.min(positions[...,1])-0.2 * (np.abs(np.min(positions[...,1])) +5)
    #         tempyx = np.max(positions[...,1]) + 0.2 * (np.abs(np.max(positions[...,1])) +5)
    #         axs.set_xlim(tempxi, tempxx)
    #         axs.set_ylim(tempyi, tempyx)

    #         #fig.tight_layout()
    # #plt.tight_layout()
    # #plt.autoscale()
    return positions


def constrains(z):
    n = z.shape[0]
    z = z - z[0,:]

    if z[1,0] < 0:
        z[:, 0] *= -1
    
    rot = torch.atan(-1 * z[1,1]/z[1,0])
    R = torch.tensor([ [torch.cos(rot), -1 * torch.sin(rot)], [torch.sin(rot), torch.cos(rot)]])

    z = torch.matmul(R, z.T)
    z = z.T
    if z.shape[1] > 2 and z[2,1] < 0:
        z[:, 1] *= -1
    
    return z
    





# Copyright © 2021 by Northwestern University.
# 
# LVGP-PyTorch is copyrighted by Northwestern University. It may be freely used 
# for educational and research purposes by  non-profit institutions and US government 
# agencies only. All other organizations may use LVGP-PyTorch for evaluation purposes 
# only, and any further uses will require prior written approval. This software may 
# not be sold or redistributed without prior written approval. Copies of the software 
# may be made by a user provided that copies are not sold or distributed, and provided 
# that copies are used under the same terms and conditions as agreed to in this 
# paragraph.
# 
# As research software, this code is provided on an "as is'' basis without warranty of 
# any kind, either expressed or implied. The downloading, or executing any part of this 
# software constitutes an implicit agreement to these terms. These terms and conditions 
# are subject to change at any time without prior notice.

import numpy as np
import io
from collections import OrderedDict
from typing import Optional,Union,List,Tuple,Dict,Callable

from .variables import Variable,NumericalVariable,CategoricalVariable,IntegerVariable

class InputSpace(object):
    '''
    Utility class containing a set of input variables along with their domains, and some
    additional methods for sampling and standardizing the inputs.

    When sampling, the inputs are always returned in array form. Categorical variables are
    encoded as ordinal integers - 0,1,...,L-1, where L is the number of levels. Numerical and 
    integer variables are scaled to the unit hypercube [0,1]. If the `logscale` attribute of 
    any such variable is `True`, the variable is first log-transformed and then scaled. 
    
    Use the `get_dict_from_array` method to get a dictionary of the actual input variable values
    for a single sample. For multiple samples, iteratively call this method over each sample.

    In array form, the variables are ordered in the order that are added to the input space object.
    Use the attributes :attr:`qual_index` and :attr:`quant_index` to obtain the list of qualitative
    or quantitative variables respectively.

    Useful attributes:
        :attr:`qual_index` (List[int])
            List containing the indices of the qualitative inputs in array form. Note that binary
            variables are not included here.
        :attr:`quant_index` (List[int])
            List containing the indices of the quantitative inputs in array form. This list also
            includes binary variables.
        :attr:`num_levels` (Dict[int,int])
            Ordered dictionary with the index of the qualitative variables as keys
            and their corresponding number of levels as values
        :attr:`n_dim` (int)
            Number of variables in the input space
    '''
    def __init__(self):
        super().__init__()

        self._variables = OrderedDict()
        self._variable_index = OrderedDict()
        self._variable_names = []
        
        self.ndim = 0
        self.quant_index = []
        self.qual_index = []
        self.num_levels = OrderedDict()
    
    def add_input(self,variable:Variable) -> None:
        '''
        Adds a variable to the input space and update the indices

        :param variable: The variable to add
        :type variable: Variable
        '''
        if not isinstance(variable,Variable):
            raise TypeError
        elif variable.name in self._variables:
            raise ValueError(
                'Variable %s is already in the space'%variable.name
            )
        
        self._variables[variable.name] = variable
        self._variable_names.append(variable.name)
        self._variable_index[variable.name] = self.ndim

        if isinstance(variable,CategoricalVariable):
            if len(variable.levels) > 2:
                # including only those non-binary categorical variables
                # it is preferable to treat binary variables as 
                # quantitative anyways
                self.qual_index.append(self.ndim)
                self.num_levels[self.ndim] = len(variable.levels)
            else:
                self.quant_index.append(self.ndim)
        else:
            self.quant_index.append(self.ndim)
        # update number of parameters
        self.ndim += 1
    
    def add_inputs(self,var_list:List[Variable]) -> None:
        '''
        Adds variables in :attr:`var_list` to the input space

        :param var_list: List of variables to add.
        :type var_list: List[Variable]
        '''
        for var in var_list:
            self.add_input(var)
    
    def get_variables(self) -> List[Variable]:
        '''
        Return the list of variables in the input space (in order)
        '''
        return list(self._variables.values())
    
    def get_variable_names(self) -> List[str]:
        '''
        Return the list of names of the variables in the input space (in order)
        '''
        return self._variables.keys()

    def get_variable_by_name(self,name:str) -> Variable:
        '''
        Gets the variable from the input space from its name. Returns `None` if
        there is no variable with the given name

        :param name: Name of the searched variable
        :type name: str 
        '''
        return self._variables.get(name)
    
    def get_variable_by_idx(self,idx:int) -> Variable:
        '''
        Gets the variable from the input space from its index in array from. Throws an
        error if index is invalid. 

        :param idx: Index of the searched variable
        :type idx: int
        '''
        return self.get_variable_by_name(
            self._variable_names[idx]
        )

    def __repr__(self):
        s = io.StringIO()
        s.write('Input space with variables:\n')
        
        if len(self._variables) > 0:
            s.write('\n')
            s.write('\n'.join([
                str(var) for var in self.get_variables()
            ]))
        return s.getvalue()
    
    def __len__(self):
        return self.ndim

    def get_dict_from_array(
        self,x:np.ndarray
    ) -> Dict[str,Union[str,float,int]]:
        '''
        Given an input sample in array form, returns a dictionary of variable-value
        pairs. The values are returned in the original scale. 

        :param x: 1D array containing the input sample in the transformed scale
        :type x: np.ndarray
        '''
        conf_dict = {}
        for i,var in enumerate(self.get_variables()):
            conf_dict[var.name] = var._transform_scalar(x[i])
        return conf_dict

    def get_array_from_dict(
        self,values: Dict[str,Union[str,float,int]]
    ) -> np.ndarray:
        '''
        Given a dictionary with the variable-value pairs, return an array with 
        the transformed values

        :param values: dictionary containing variable value pairs. This needs to 
            contain all variables in the input space - missing variables will raise
            errors
        :type values: Dict[str,Union[str,float,int]]
        '''
        out = np.zeros(self.ndim)
        for i,var in enumerate(self.get_variables()):
            out[i] = var._inverse_transform_scalar(values[var.name])
        return out

    def random_sample(
        self,
        rng:Optional[Union[int,np.random.RandomState]]=None,
        size:int = 1,
    ) -> np.ndarray:
        '''
        Returns samples drawn uniformly at random in array form

        :param rng: The seed of the psuedo random number generator. This can either 
            be a RandomState instance or an integer that can be used to initialize one.
        :type rng: Union[int,np.random.RandomState], optional

        :param size: Number of samples to draw
        :type size: int

        :returns: A :attr:`size` x :attr:`ndim` array with the different values in 
            the transformed scale
        :rtype: np.ndarray
        '''
        rng = _check_random_state(rng)
        out = np.zeros((size,self.ndim))
        for i,var in enumerate(self.get_variables()):
            out[:,i] = var.sample(rng,size)
        
        return out

    def latinhypercube_sample(
        self,
        rng:Optional[Union[int,np.random.RandomState]]=None,
        size:int=1,
    ) -> np.ndarray:
        '''
        Returns samples drawn from a "latin-hypercube" design in array form. 

        :param rng:  The seed of the psuedo random number generator. This can either 
            be a RandomState instance or an integer that can be used to initialize one.
        :type rng: Union[int,np.random.RandomState], optional
        
        :param size: Number of samples to draw
        :type size: int

        :returns: A :attr:`size` x :attr:`ndim` array with the different values in 
            the transformed scale
        :rtype: np.ndarray
        '''
        rng = _check_random_state(rng)
        out = np.zeros((size,self.ndim))
        if len(self.quant_index) > 0:
            out[:,self.quant_index] = _latinhypercube_sample(
                rng,len(self.quant_index),size
            )
        
        if len(self.qual_index) > 0:
            for idx in self.qual_index:
                out[:,idx] = self.get_variable_by_idx(idx).stratified_sample(rng,size)
        
        return out

def _check_random_state(rng):
    if rng is None or rng is np.random:
        return np.random.mtrand._rand
    elif isinstance(rng,int):
        return np.random.RandomState(rng)
    elif isinstance(rng,np.random.RandomState):
        return rng
    else:
        raise ValueError(
            "%r cannot be used to seed a numpy.random.RandomState instance" % rng
        )


def _latinhypercube_sample(
    rng:Union[int,Callable,np.random.RandomState],
    ndim:int,size:int)->np.ndarray:
    # generate row and column grids
    grid_bounds = np.stack([np.linspace(0., 1., size+1) \
                            for i in np.arange(ndim)],axis=1)
    grid_lower = grid_bounds[:-1,:]
    grid_upper = grid_bounds[1:,:]
    
    # generate 
    grid = grid_lower + (grid_upper-grid_lower)*rng.rand(size,ndim)
    
    # shuffle and return
    for i in range(ndim):
        rng.shuffle(grid[:,i])
    
    return grid


import numpy as np
import torch 


def interval_score_function(Yu, Yl, Y, alpha = 0.05):
    out = Yu - Yl
    out += (Y > Yu).to(torch.int64) * 2/alpha * (Y- Yu)
    out += (Y < Yl).to(torch.int64) * 2/alpha * (Yl - Y)
    
    accuracy = torch.sum(out > (Yu - Yl)).to(torch.float64)
    accuracy /= len(out)

    return torch.mean(out), accuracy



# Copyright © 2023, Dr. Bostanabad's research group at the University of California, Irvine.
# 
# GP+ Intellectual Property Notice:
# 
# The software known as GP+ is the proprietary material of Dr. Bostanabad's research group at the University of California, Irvine. 
# Non-profit academic institutions and U.S. government agencies may utilize this software exclusively for educational and research endeavors. 
# All other entities are granted permission for evaluation purposes solely; any additional utilization demands prior written consent from the appropriate authority. 
# The direct sale or redistribution of this software, in any form, without explicit written authorization is strictly prohibited. 
# Users are permitted to make duplicate copies of the software, contingent upon the assurance that no copies are sold or redistributed and they adhere to the stipulated terms herein.
# 
# Being academic research software, GP+ is provided on an "as is" m_gp, devoid of warranties, whether explicit or implicit. 
# The act of downloading or executing any segment of this software inherently signifies compliance with these terms. 
# The developers reserve the right to modify these terms and conditions without prior intimation at any juncture.

from distutils.log import error
from turtle import forward
from gpytorch.likelihoods import _GaussianLikelihoodBase
from gpytorch.likelihoods.noise_models import _HomoskedasticNoiseBase
from sklearn.covariance import log_likelihood
import torch
from typing import Any, Optional
from gpytorch.distributions import MultivariateNormal
from gpytorch.lazy import DiagLazyTensor, ConstantDiagLazyTensor
from torch import Tensor

class Multifidelity_likelihood(_GaussianLikelihoodBase):

    def __init__(self, fidel_indices: Tensor,noise_indices: list = [1], noise_prior = None, noise_constraint = None, 
        learn_additional_noise = False, batch_shape = torch.Size(), **kwargs) -> None:
        # num_noises = len(noise_indices)
        num_noises = len(noise_indices)
        noise_covar = Multifidelity_noise(noise_prior=noise_prior, 
        noise_constraint=noise_constraint, batch_shape=batch_shape, num_noises = num_noises)
        super().__init__(noise_covar = noise_covar)


        self.fidel_indices = fidel_indices
        self.noise_indices = noise_indices

    @property
    def noise(self) -> Tensor:
        return self.noise_covar.noise

    @noise.setter
    def noise(self, value: Tensor) -> None:
        self.noise_covar.initialize(noise=value)

    @property
    def raw_noise(self) -> Tensor:
        return self.noise_covar.raw_noise

    @raw_noise.setter
    def raw_noise(self, value: Tensor) -> None:
        self.noise_covar.initialize(raw_noise=value)


    def _shaped_noise_covar(self, base_shape: torch.Size, *params: Any, **kwargs: Any):
        # This runs the forward method in noise class
        # Shape is not used any more.
        return self.noise_covar(*params, fidel_indices = self.fidel_indices, noise_indices = self.noise_indices)
    

    def marginal(self, function_dist: MultivariateNormal, *params: Any, **kwargs: Any) -> MultivariateNormal:
        mean, covar = function_dist.mean, function_dist.lazy_covariance_matrix
        noise_covar = self._shaped_noise_covar(mean.shape, *params, **kwargs)
        full_covar = covar + noise_covar
        return function_dist.__class__(mean, full_covar)





class Multifidelity_noise(_HomoskedasticNoiseBase):
    def __init__(self, noise_prior=None, noise_constraint=None, batch_shape= torch.Size(), num_noises=1):
        super().__init__(noise_prior, noise_constraint, batch_shape, num_tasks= num_noises)


    def forward(self, *params: Any, shape: Optional[torch.Size] = None, 
        fidel_indices: Tensor, noise_indices: list, **kwargs: Any) -> DiagLazyTensor:
        
        """_summary_
        
        Indices are very important here and is coming from input and shows the multifidelity level of each input data

        Raises:
            ValueError: _description_
        """

        if len(fidel_indices) ==0 or fidel_indices is None:
            raise ValueError('You need to specify a list of indices for noise such as [1,3]')
        # This contains a list of diagonal matrices with defined noise. Crates [batch * 1 * noise_size * n * n]
        covar = super().forward(*params, shape= fidel_indices.shape, **kwargs)

        if covar.dim() > 2:
            if covar.shape[1] is not len(noise_indices):
                raise ValueError('Something is wrong, number of noise and indices are not the same')

        if covar.dim() == 4: # no batch
            covar = covar.squeeze(0)
        elif covar.dim() == 5: # for batch
            covar = covar.squeeze(1) 


        # This part is for categorical_indices
        temp = ConstantDiagLazyTensor(torch.tensor([0.0]), len(fidel_indices))
        temp = temp.to(dtype=covar.dtype, device=covar.device)
        for i in range(len(noise_indices)):
            if i ==0:
                diag = DiagLazyTensor( (fidel_indices == noise_indices[i]) )#.type(torch.int32)
                if covar.dim() == 4: # batch
                    temp += diag * covar[:,i,...]
                
                elif covar.dim() == 3: # 
                    temp += diag * covar[i,...]

                elif covar.dim() == 2:
                    temp += diag * covar
                else:
                    raise ValueError('Covar is 1D? why?')
            else :
                diag = DiagLazyTensor( (fidel_indices == noise_indices[i]).type(torch.int32) )
                if covar.dim() == 4: # batch
                    temp += diag * covar[:,i,...]
                
                elif covar.dim() == 3: # 
                    temp += diag * covar[i,...]

                elif covar.dim() == 2:
                    temp += diag * covar
                else:
                    raise ValueError('Covar is 1D? why?')
        

        

        return temp


'''
if __name__ == '__main__':
    multi_likelihood = Multifidelity_likelihood(num_noises=3)
    multi_noise = Multifidelity_noise(num_noises=2)
    print(multi_likelihood)
    fidel_indices = torch.tensor([1, 3, 2, 3, 2, 1, 3])
    noise_indices = [1, 3]
    covar = multi_noise(fidel_indices = fidel_indices, noise_indices = noise_indices)
    aa = 1
'''


import numpy as np
import matplotlib.pyplot as plt
import torch
import matplotlib.cm as cm


def plot_sep(type ,positions, levels, perm, constraints_flag = True, suptitle= None):
    if positions.ndim > 2:
        positions = positions.mean(axis = 0)
    else:
        positions = positions.detach()

    if positions.ndim > 2:
        positions = positions.mean(axis = 0) 

    # applying the constrains
    if constraints_flag:
        positions = constrains(positions)


    positions = positions.cpu().detach().numpy()
    legend=[ 'HF', 'LF1','LF2','LF3','LF4','LF5','LF6','LF7','LF8','LF9','LF10','LF11','LF12','LF13','LF14','LF15','LF16','LF17','LF18','LF19','LF20']

    # plt.rcParams.update({'font.size': 14})
    plt.rcParams['font.family'] = 'Times New Roman'
    plt.rcParams['font.size'] = 18
    plt.rcParams['figure.dpi']=150
    #colors = {0:'blue', 1:'r', 2:'g', 3:'c', 4:'m', 5:'k', 6:'y'}
    # tab20 = plt.get_cmap('tab10')
    colors_base = ['deeppink', 'gold', 'darkorange', 'gray', 'orangered', 'blue']
    cmap = cm.get_cmap('tab20', 20)  # Get a colormap from matplotlib, 'tab20' has nice distinct colors
    colors = colors_base  + [cmap(i) for i in range(len(colors_base), 20)]
    marker = ['X','o','s',"v", 'p', 'v', '^', '<', '>','*', 'h', 'H', '+', 'x', 'X', 'D', 'd', '|', '_','1', '2', '3', '4',]
    # loop over the number of variables
    if len(levels) > 1:
        fig,axs = plt.subplots(1, len(levels),figsize=(10,4))
        for j in range(len(levels)):
            for i in range(levels[j]):
                index = torch.where(perm[:,j] == i) 
                if i<=10:
                    fontsize=5
                    s_size=100
                else:
                    fontsize=5
                    s_size=100
                if  type=='mf':
                    axs[j].scatter(positions[index][...,0], positions[index][...,1], label = legend[i], color = colors[i], marker=marker[i],s=s_size,alpha=.6)#marker=r'$\clubsuit$'
                    plt.xlabel(r'$z_1$',labelpad=0,rotation=0,fontsize=20)
                    plt.ylabel(r'$z_2$',labelpad=10,rotation=0, fontsize=20)
                
                elif type=='cat':
                    axs[j].scatter(positions[index][...,0], positions[index][...,1], label = 'level' + str(i+1), color = colors[i], marker=marker[i],s=s_size,alpha=.6)#marker=r'$\clubsuit$'
                    axs[j].set_xlabel(r'$h_1$', labelpad=0, fontsize=20)
                    axs[j].set_ylabel(r'$h_2$', labelpad=5,fontsize=20)
                else:
                    raise ValueError("type should be either 'mf' or cat:")
                # plt.tight_layout()
                # axs[j].legend()
                axs[j].legend(loc='upper right', fontsize='xx-small')

                tempxi = np.min(positions[...,0])-0.2 * (np.abs(np.min(positions[...,0])) +5)
                tempxx = np.max(positions[...,0]) + 0.2 * (np.abs(np.max(positions[...,0])) +5)
                tempyi = np.min(positions[...,1])-0.2 * (np.abs(np.min(positions[...,1])) +5)
                tempyx = np.max(positions[...,1]) + 0.2 * (np.abs(np.max(positions[...,1])) +5)
                axs[j].set_xlim(tempxi, tempxx)
                axs[j].set_ylim(tempyi, tempyx)
    else:
        fig,axs = plt.subplots(1, len(levels),figsize=(6,4))
        for j in range(len(levels)):
            for i in range(levels[j]):
                index = torch.where(perm[:,j] == i) 
                if i<=7:
                    fontsize=20
                    s_size=200
                else:
                    fontsize=20
                    s_size=100
                if  type=='mf':
                    axs.scatter(positions[index][...,0], positions[index][...,1], label = legend[i], color = colors[i], marker=marker[i],s=s_size,alpha=.6)#marker=r'$\clubsuit$'
                    plt.xlabel(r'$z_1$',labelpad=0,rotation=0, fontsize=fontsize)
                    plt.ylabel(r'$z_2$',labelpad=10,rotation=0, fontsize=fontsize)
                
                elif type=='cat':
                    axs.scatter(positions[index][...,0], positions[index][...,1], label = 'level' + str(i+1), color = colors[i], marker=marker[i],s=s_size,alpha=.6)#marker=r'$\clubsuit$'
                    axs.set_xlabel(r'$h_1$', labelpad=0, fontsize=fontsize)
                    axs.set_ylabel(r'$h_2$', labelpad=5,fontsize=fontsize)
                else:
                    raise ValueError("type should be either 'mf' or cat:")
                # plt.tight_layout()
                # axs[j].legend()
                axs.legend(loc='upper right', fontsize='xx-small')

                tempxi = np.min(positions[...,0])-0.2 * (np.abs(np.min(positions[...,0])) +5)
                tempxx = np.max(positions[...,0]) + 0.2 * (np.abs(np.max(positions[...,0])) +5)
                tempyi = np.min(positions[...,1])-0.2 * (np.abs(np.min(positions[...,1])) +5)
                tempyx = np.max(positions[...,1]) + 0.2 * (np.abs(np.max(positions[...,1])) +5)
                axs.set_xlim(tempxi, tempxx)
                axs.set_ylim(tempyi, tempyx)
    fig.tight_layout()

    plt.show()
    if suptitle is not None:
        plt.suptitle(suptitle,fontsize=18)


def constrains(z):
    n = z.shape[0]
    z = z - z[0,:]

    if z[1,0] < 0:
        z[:, 0] *= -1
    
    rot = torch.atan(-1 * z[1,1]/z[1,0])
    R = torch.tensor([ [torch.cos(rot), -1 * torch.sin(rot)], [torch.sin(rot), torch.cos(rot)]])

    z = torch.matmul(R, z.T)
    z = z.T
    if z.shape[1] > 2 and z[2,1] < 0:
        z[:, 1] *= -1
    
    return z
    





# data_reader.py in the gpplus package
import pandas as pd
import pkg_resources

def data_reader(file_name):
    # Construct the resource name
    resource_name = pkg_resources.resource_filename('gpplus', 'datasets/' + file_name)

    # Determine the file type and read accordingly
    if file_name.endswith('.csv'):
        return pd.read_csv(resource_name)
    elif file_name.endswith('.xlsx'):
        return pd.read_excel(resource_name)
    else:
        raise ValueError("Unsupported file format: Only CSV and Excel files are supported.")


from gpplus.preprocessing import standard 
from gpplus.preprocessing import setlevels
from sklearn.model_selection import train_test_split
import torch
import warnings

def train_test_split_normalizeX(
    X,
    y,
    test_size=None,
    shuffle=True,
    stratify=None,
    qual_dict = {},
    random_state=1,
    return_mean_std = False,
    set_levels = False
):
    # Finding the quant index from qual index
    qual_index = list(qual_dict.keys())
    # all_index = set(range(X.shape[-1]))
    # if len(Calibration_index)>0:
    #     quant_index = list(set(quant_index).difference(Calibration_index))

    if set_levels:
        # This will assign levels to categorical evenif the levels are strings
        X = setlevels(X, qual_index = qual_index)
    # Split test and train
    Xtrain, Xtest, ytrain, ytest = train_test_split(X, y, 
        test_size= test_size, shuffle= shuffle,random_state=random_state, stratify=stratify)
    # Standard
    Xtrain, Xtest, mean_train, std_train = standard(Xtrain = Xtrain, 
        qual_index = qual_dict, Xtest = Xtest)

    variables = {'Xtrain': Xtrain, 'Xtest': Xtest, 'ytrain': ytrain, 'ytest': ytest}
    for var_name, var in variables.items():
        if not isinstance(var, torch.Tensor):
            original_type = type(var).__name__
            # warnings.warn(f"'{var_name}' was not a torch.Tensor (type: {original_type}). It is converted to torch.Tensor to proceed with the preprocessing.")
            variables[var_name] = torch.tensor(var)

    Xtrain, Xtest, ytrain, ytest = variables['Xtrain'], variables['Xtest'], variables['ytrain'], variables['ytest']



    if return_mean_std:
        return Xtrain, Xtest, ytrain, ytest, mean_train, std_train

    return Xtrain, Xtest, ytrain, ytest


from gpytorch.kernels import MaternKernel
from functools import partialmethod

class Matern32Kernel(MaternKernel):
    __init__ = partialmethod(MaternKernel.__init__,nu=1.5)

class Matern52Kernel(MaternKernel):
    __init__ = partialmethod(MaternKernel.__init__,nu=2.5)


# Copyright © 2021 by Northwestern University.
# 
# LVGP-PyTorch is copyrighted by Northwestern University. It may be freely used 
# for educational and research purposes by  non-profit institutions and US government 
# agencies only. All other organizations may use LVGP-PyTorch for evaluation purposes 
# only, and any further uses will require prior written approval. This software may 
# not be sold or redistributed without prior written approval. Copies of the software 
# may be made by a user provided that copies are not sold or distributed, and provided 
# that copies are used under the same terms and conditions as agreed to in this 
# paragraph.
# 
# As research software, this code is provided on an "as is'' basis without warranty of 
# any kind, either expressed or implied. The downloading, or executing any part of this 
# software constitutes an implicit agreement to these terms. These terms and conditions 
# are subject to change at any time without prior notice.

import torch

softplus = torch.nn.Softplus()

def inv_softplus(x):
    return x + torch.log(-torch.expm1(-x))


# Copyright © 2023, Dr. Bostanabad's research group at the University of California, Irvine.
# 
# GP+ Intellectual Property Notice:
# 
# The software known as GP+ is the proprietary material of Dr. Bostanabad's research group at the University of California, Irvine. 
# Non-profit academic institutions and U.S. government agencies may utilize this software exclusively for educational and research endeavors. 
# All other entities are granted permission for evaluation purposes solely; any additional utilization demands prior written consent from the appropriate authority. 
# The direct sale or redistribution of this software, in any form, without explicit written authorization is strictly prohibited. 
# Users are permitted to make duplicate copies of the software, contingent upon the assurance that no copies are sold or redistributed and they adhere to the stipulated terms herein.
# 
# Being academic research software, GP+ is provided on an "as is" m_gp, devoid of warranties, whether explicit or implicit. 
# The act of downloading or executing any segment of this software inherently signifies compliance with these terms. 
# The developers reserve the right to modify these terms and conditions without prior intimation at any juncture.

import torch
import gpytorch
import math
from gpytorch.models import ExactGP
from gpytorch import settings as gptsettings
from gpytorch.priors import NormalPrior,LogNormalPrior
from gpytorch.constraints import GreaterThan,Positive
from gpytorch.distributions import MultivariateNormal
from .. import kernels
from ..priors import LogHalfHorseshoePrior,MollifiedUniformPrior
from ..utils.transforms import softplus,inv_softplus
from typing import List,Tuple,Union
from gpplus.likelihoods_noise.multifidelity import Multifidelity_likelihood
from botorch.models.utils import gpt_posterior_settings
from botorch.posteriors.gpytorch import GPyTorchPosterior
from botorch.models.gpytorch import BatchedMultiOutputGPyTorchModel, GPyTorchModel
from botorch import settings
from botorch.models.utils import fantasize as fantasize_flag, validate_input_scaling
from botorch.sampling.samplers import MCSampler
from torch import Tensor
from typing import Any, Dict, List, Optional, Union


class GPR(ExactGP, GPyTorchModel):
    def __init__(
        self,
        train_x:torch.Tensor,
        train_y:torch.Tensor,
        correlation_kernel,
        noise_indices:List[int],
        fix_noise:bool=False,
        fix_noise_val:float=1e-5,
        lb_noise:float=1e-12,
    ) -> None:
        # check inputs
        if not torch.is_tensor(train_x):
            raise RuntimeError("'train_x' must be a tensor")
        if not torch.is_tensor(train_y):
            raise RuntimeError("'train_y' must be a tensor")

        if train_x.shape[0] != train_y.shape[0]:
            raise RuntimeError("Inputs and output have different number of observations")
        
        # initializing likelihood
        noise_constraint=GreaterThan(lb_noise,transform=torch.exp,inv_transform=torch.log)
        
        if len(noise_indices) == 0:

            likelihood = gpytorch.likelihoods.GaussianLikelihood(noise_constraint=noise_constraint)
        else:

            likelihood = Multifidelity_likelihood(noise_constraint=noise_constraint, noise_indices=noise_indices, fidel_indices=train_x[:,-1])
        y_min= train_y.min()
        y_std= train_y.max()-train_y.min()
        train_y_sc = (train_y-y_min)/y_std

        ExactGP.__init__(self, train_x,train_y_sc, likelihood)
        
        # registering mean and std of the raw response
        self.register_buffer('y_min',y_min)
        self.register_buffer('y_std',y_std)
        self.register_buffer('y_scaled',train_y_sc)

        self._num_outputs = 1

        # initializing and fixing noise
        # if noise is not None:
        #     self.likelihood.initialize(noise=noise)
        
        self.likelihood.register_prior('noise_prior',LogHalfHorseshoePrior(0.01,lb_noise),'raw_noise')
        if fix_noise:
            self.likelihood.raw_noise.requires_grad_(False)
            self.likelihood.noise_covar.noise =torch.tensor(fix_noise_val)

        if isinstance(correlation_kernel,str):
            try:
                correlation_kernel_class = getattr(kernels,correlation_kernel)
                correlation_kernel = correlation_kernel_class(
                    ard_num_dims = self.train_inputs[0].size(1),
                    lengthscale_constraint=Positive(transform=torch.exp,inv_transform=torch.log),
                )
                correlation_kernel.register_prior(
                    'lengthscale_prior',MollifiedUniformPrior(math.log(0.1),math.log(10)),'raw_lengthscale'
                )
            except:
                raise RuntimeError(
                    "%s not an allowed kernel" % correlation_kernel
                )
        elif not isinstance(correlation_kernel,gpytorch.kernels.Kernel):
            raise RuntimeError(
                "specified correlation kernel is not a `gpytorch.kernels.Kernel` instance"
            )

        self.covar_module = kernels.ScaleKernel(
            base_kernel = correlation_kernel,
            outputscale_constraint=Positive(transform=softplus,inv_transform=inv_softplus),
        )
        # register priors
        self.covar_module.register_prior(
            'outputscale_prior',LogNormalPrior(1e-6,1.),'outputscale'
        )
    
    def forward(self,x:torch.Tensor)->MultivariateNormal:
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x,covar_x)
    
    def predict(
        self,x:torch.Tensor,return_std:bool=False,include_noise:bool=False
    )-> Union[torch.Tensor,Tuple[torch.Tensor]]:

        self.eval()
        with gptsettings.fast_computations(log_prob=False):
            # determine if batched or not
            ndim = self.train_targets.ndim
            if ndim == 1:
                output = self(x)
            else:
                # for batched GPs 
                num_samples = self.train_targets.shape[0]
                output = self(x.unsqueeze(0).repeat(num_samples,1,1))
            self.fidel_indices=x[:,-1]
            if return_std and include_noise:
                # x=self.fidel_indices
                self.likelihood.fidel_indices=x[:,-1]   ### the fidelity indaces are alrrady set for training data, here we update them test data since 
                output = self.likelihood(output)

            out_mean = self.y_min + self.y_std*output.mean
            
            # standard deviation may not always be needed
            if return_std:
                out_std = output.variance.sqrt()*self.y_std
                return out_mean,out_std

            return out_mean

    def posterior(
        self,
        X,
        output_indices = None,
        observation_noise= True,
        posterior_transform= None,
        **kwargs,
    ):

        self.eval()
        with gpt_posterior_settings() and gptsettings.fast_computations(log_prob=False):
    
            if observation_noise:
                return GPyTorchPosterior(mvn = self.likelihood(self(X.double())))
            else:
                return GPyTorchPosterior(mvn = self(X.double()))
    
    def reset_parameters(self) -> None:
        """Reset parameters by sampling from prior
        """
        for _,module,prior,closure,setting_closure in self.named_priors():
            if not closure(module).requires_grad:
                continue
            setting_closure(module,prior.expand(closure(module).shape).sample().to(**self.tkwargs))


    def fantasize(
            self,
            X: Tensor,
            sampler: MCSampler,
            observation_noise: Union[bool, Tensor] = True,
            **kwargs: Any,
        ):
            r"""Constructs a fantasy model using a specified procedure.

            This method constructs a fantasy model by following these steps:
            1. Compute the model's posterior at `X`. If `observation_noise=True`, the posterior
            includes observation noise, which is determined as the mean of the observation
            noise in the training data. If `observation_noise` is a Tensor, it is used directly
            as the observation noise.
            2. Sample from this posterior using the provided `sampler` to create "fake" observations.
            3. Update (condition) the model with these new fake observations.

            Args:
                X: A Tensor of dimensions `batch_shape x n' x d`, where `d` represents the feature
                space dimension, `n'` is the number of points per batch, and `batch_shape`
                is the batch shape. This batch shape must be compatible with the model's
                existing batch shape.
                sampler: A sampler used for drawing samples from the model's posterior at `X`.
                observation_noise: A boolean or a Tensor. If True, the mean of the observation
                                noise from the training data is used in the posterior. If a
                                Tensor, it specifies the observation noise directly.

            Returns:
                A fantasy model, updated based on the sampled fake observations.
            """
            propagate_grads = kwargs.pop("propagate_grads", False)
            with fantasize_flag():
                with settings.propagate_grads(propagate_grads):
                    post_X = self.posterior(
                        X, observation_noise=observation_noise, **kwargs
                    )
                Y_fantasized = sampler(post_X)  # num_fantasies x batch_shape x n' x m
                # Use the mean of the previous noise values (TODO: be smarter here).
                # noise should be batch_shape x q x m when X is batch_shape x q x d, and
                # Y_fantasized is num_fantasies x batch_shape x q x m.
                noise_shape = Y_fantasized.shape[1:]
                noise = self.likelihood.noise.mean().expand(noise_shape)
                return self.condition_on_observations(
                    X=self.transform_inputs(X), Y=Y_fantasized, noise=noise
                )



from gpytorch.kernels import RBFKernel
from gpytorch.functions import RBFCovariance
from gpytorch.settings import trace_mode
import torch

def postprocess_rbf(dist_mat):
    return dist_mat.div_(-1).exp_()


class wighted_RBF(RBFKernel):
    """_summary_

    Args:
        RBFKernel (_type_): _description_
    """
    has_lengthscale = True

    def forward(self, x1, x2, diag=False, **params):
        if (
            x1.requires_grad
            or x2.requires_grad
            or (self.ard_num_dims is not None and self.ard_num_dims > 1)
            or diag
            or params.get("last_dim_is_batch", False)
            or trace_mode.on()
        ):
            ten_power_omega_sqrt = self.lengthscale.sqrt()
            x1_ = x1.mul(ten_power_omega_sqrt)
            #### Amin Added this
            x2_ = x2.mul(ten_power_omega_sqrt)
            AA=self.covar_dist(
                x1_, x2_, square_dist=True, diag=diag, dist_postprocess_func=postprocess_rbf, postprocess=True, **params
            )
            # EQ=1/ten_power_omega_sqrt+0*AA
            weight_all=1
            weight_H=1

            EQ=weight_all+0*AA
            WE=EQ[0,0]
            N=(-WE+weight_H)*torch.diag(x1.reshape(1,-1)[0])
            return EQ+N
        return RBFCovariance.apply(
            x1,
            x2,
            self.lengthscale,
            lambda x1, x2: self.covar_dist(
                x1, x2, square_dist=True, diag=False, dist_postprocess_func=postprocess_rbf, postprocess=False, **params
            ),
        )


import numpy as np
from scipy.stats.qmc import Sobol, scale
import math
import torch
from pyDOE import lhs

################################## Wing #########################################
def wing(n=100, X = None, 
    fidelity = 0,noise_std = 0.0, random_state = None, shuffle = True):
    if random_state is not None:
        np.random.seed(random_state)

    dx = 10
    l_bound = [150, 220, 6, -10, 16, 0.5, 0.08, 2.5, 1700, 0.025]
    u_bound = [200, 300, 10, 10, 45, 1, 0.18, 6, 2500, 0.08]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)
    Sw = X[..., 0]
    Wfw = X[..., 1]
    A = X[..., 2]
    Gama = X[..., 3] * (np.pi/180.0)
    q = X[..., 4]
    lamb = X[..., 5]
    tc = X[..., 6]
    Nz = X[..., 7]
    Wdg = X[..., 8]
    Wp = X[..., 9]
    # This is the output

    if fidelity == 0:
        y = 0.036 * Sw**0.758 * Wfw**0.0035 * (A/(np.cos(Gama)) ** 2) ** 0.6 * \
            q**0.006*lamb**0.04 * ((100 * tc)/(np.cos(Gama)))**(-0.3) *\
            (Nz * Wdg) ** 0.49 + Sw * Wp
    elif fidelity == 1:
        y = 0.036 * Sw**0.758 * Wfw**0.0035 * (A/(np.cos(Gama)) ** 2) ** 0.6 * \
            q**0.006*lamb**0.04 * ((100 * tc)/(np.cos(Gama)))**(-0.3)\
            * (Nz * Wdg) ** 0.49 + 1 * Wp 

    elif fidelity == 2:
        y = 0.036 * Sw**0.8 * Wfw**0.0035 * (A/(np.cos(Gama)) ** 2) ** 0.6 * \
            q**0.006*lamb**0.04 * ((100 * tc)/(np.cos(Gama)))**(-0.3)\
            * (Nz * Wdg) ** 0.49 + 1 * Wp


    elif fidelity == 3:
        y = 0.036 * Sw**0.9 * Wfw**0.0035 * (A/(np.cos(Gama)) ** 2) ** 0.6 * \
            q**0.006*lamb**0.04 * ((100 * tc)/(np.cos(Gama)))**(-0.3)\
            * (Nz * Wdg) ** 0.49 + 0 * Wp
    else:
        raise ValueError('only 4 fidelities of 0,1,2,3 have been implemented ')

    if X is None:
        if shuffle:
            index = np.random.randint(0, len(y), size = len(y))
            X = X[index,...]
            y = y[index]

    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y


################################# Multi-fidelity ####################################
def multi_fidelity_wing(X = None,
    n={'0': 50, '1': 100, '2': 100, '3': 100},
    noise_std={'0': 0.0, '1': 0.0, '2': 0.0, '3': 0.0},
    random_state = None, shuffle = True):
    if X is None:
        X_list = []
        y_list = []
        for level, num in n.items():
            if level  in ['0','1','2','3'] and num > 0:
                X, y = wing(n=num, fidelity = int(level), 
                    noise_std= noise_std[level], random_state = random_state)
                X = np.hstack([X, np.ones(num).reshape(-1, 1) * float(level)])
                X_list.append(X)
                y_list.append(y)
            else:
                raise ValueError('Wrong label, should be h, l1, l2 or l3')
        return np.vstack([*X_list]), np.hstack(y_list)
    else:
        fidelity = [i for i in n.keys()]
        y_list = []
        if type(X) == np.ndarray:
            X = torch.tensor(X)
        for f in fidelity:
            # Find the index for each fidelity then call the wing
            index = [i[0] for i in torch.argwhere(X[...,-1]== int(f))]
            y_list.append(wing(X=X[index, 0:-1], fidelity=int(f),  
                noise_std= noise_std[f]))
        return torch.tensor(np.hstack(y_list))



def multi_fidelity_wing_value(input):
    y_list = []
    for value in input:
        if value[-1] == 0.0:
            y_list.append(wing(X=value))
        elif value[-1] == 1.0:
            y_list.append(wing(X=value))
        elif value[-1] == 2.0:
            y_list.append(wing(X=value))
        elif value[-1] == 3.0:
            y_list.append(wing(X=value))
        else:
            raise ValueError('Wrong label, should be 0, 1, 2 or 3')
    return torch.tensor(np.hstack(y_list))


def Augmented_branin(input, negate = True, mapping = None):

    X = input.clone()

    if mapping is not None:
        X[..., 2] = torch.tensor(list(map(lambda x: mapping[str(float(x))], X[..., 2]))).to(X)

    t1 = (
        X[..., 1]
        - (5.1 / (4 * math.pi ** 2) - 0.1 * (1 - X[..., 2])) * X[..., 0] ** 2
        + 5 / math.pi * X[..., 0]
        - 6
    )
    t2 = 10 * (1 - 1 / (8 * math.pi)) * torch.cos(X[..., 0])
    return -(t1 ** 2 + t2 + 10) if negate else (t1 ** 2 + t2 + 10)


def Borehole_MF_BO(init_data,x,var = (0, 0, 0, 0,0)):
    # Define functions
    def y_h(x,noise_std=4):
        Tu, Hu, Hl, r, rw, Tl, L, Kw = [x[:, i] for i in range(8)]
        Numer = 2*np.pi*Tu
        y= (Numer*(Hu-Hl))/(np.log(r/rw)*(1+((2*L*Tu)/(np.log(r/rw)*(rw**2)*Kw))+(Tu/Tl)))
        if noise_std>0:
            return y + np.random.randn(*y.shape) * noise_std
        else:
            return y
    def y_l1(x):
        Tu, Hu, Hl, r, rw, Tl, L, Kw = [x[:, i] for i in range(8)]
        Numer = 2*np.pi*Tu
        return (Numer*(Hu-.8*Hl))/(np.log(r/rw)*(1+((1*L*Tu)/(np.log(r/rw)*(rw**2)*Kw))+(Tu/Tl)))
    def y_l2(x):
        Tu, Hu, Hl, r, rw, Tl, L, Kw = [x[:, i] for i in range(8)]
        Numer = 2*np.pi*Tu
        return (Numer*(Hu-Hl))/(np.log(r/rw)*(1+((8*L*Tu)/(np.log(r/rw)*(rw**2)*Kw))+.75*(Tu/Tl)))
    def y_l3(x):
        Tu, Hu, Hl, r, rw, Tl, L, Kw = [x[:, i] for i in range(8)]
        Numer = 2*np.pi*Tu
        return (Numer*(1.09*Hu-Hl))/(np.log(4*r/rw)*(1+((3*L*Tu)/(np.log(r/rw)*(rw**2)*Kw))+(Tu/Tl)))
    def y_l4(x):
        Tu, Hu, Hl, r, rw, Tl, L, Kw = [x[:, i] for i in range(8)]
        Numer = 2*np.pi*Tu
        return (Numer*(1.05*Hu-Hl))/(np.log(2*r/rw)*(1+((3*L*Tu)/(np.log(r/rw)*(rw**2)*Kw))+(Tu/Tl)))
    if init_data:
        # Set parameters
        n_train = tuple(x.values())
        dx = 9
        dnum = 8
        dy = 1
        # dt is a list whose indices correspond to the categorical variable
        # and whose entries correspond to the number of levels for that variable
        dt = (5,)
        dsource = (4,)
        num_idx = np.array(list(range(0, 8)))
        source_idx = np.array([8])
        MIN = (100, 990, 700, 100, .05, 10, 1000, 6000)
        MAX = (1000, 1110, 820, 10000, .15, 500, 2000, 12000)
        RANGE = [MAX[i] - MIN[i] for i in range(dnum)]
        y_list = [y_h, y_l1, y_l2, y_l3,y_l4]
        y = lambda x, t: y_list[t](x)
        # Preallocate data arrays and indices
        # First, get the total number of input points for each set
        train_tot = sum(n_train)
        x_train = np.empty([train_tot, dx])
        y_train = np.empty([train_tot, dy])
        # Initialize the indices for training
        train_start = 0
        # Generate data
        for i in range(dt[0]):
            # Set end indices
            train_end = train_start + n_train[i]
            n_train_temp = n_train[i]
            x_train_temp = lhs(dnum, samples = n_train_temp) * RANGE + MIN
            y_train_temp = y(x_train_temp, i).reshape((n_train_temp, dy))
            y_train_temp = y_train_temp + np.sqrt(var[i])*np.random.standard_normal(size=y_train_temp.shape)
            # Store the data
            x_train[train_start : train_end, 0:dnum] = x_train_temp
            x_train[train_start : train_end, dnum:dnum+source_idx[0]] = i*np.ones((np.shape(x_train_temp)[0], 1))
            y_train[train_start : train_end, :] = y_train_temp
            # Update indices
            train_start = train_end
    #(dx, dnum, dsource, dt, dy, num_idx, source_idx, min_x, max_x, min_y, max_y, x_train, y_train, x_val, y_val, x_test, y_test)
        return x_train , y_train
    else:
        if type(x) == torch.Tensor:
            input_copy = x.clone()
        elif type(x) == np.ndarray:
            input_copy = np.copy(x)
        y_list = []
        for X in input_copy:
            if X[-1].numpy() == 0.0:
                y_list.append(y_h(X[0:-1].unsqueeze(0), noise_std=2)) #, noise_std=3.31
            elif X[-1].numpy() == 1.0:
                y_list.append(y_l1(X[0:-1].unsqueeze(0)))
            elif X[-1].numpy() == 2.0:
                y_list.append(y_l2(X[0:-1].unsqueeze(0)))
            elif X[-1] == 3.0:
                y_list.append(y_l3(X[0:-1].unsqueeze(0)))
            elif X[-1]==4.0:
                y_list.append(y_l4(X[0:-1].unsqueeze(0)))
            else:
                raise ValueError('Wrong label, should be h, l1, l2 or l3')
        return torch.tensor(np.hstack(y_list))
